#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import (
    absolute_import, unicode_literals, division, print_function
    )

from astropy import units as apu
import numpy as np
from .. import conversions as cnv
from .. import utils


__all__ = [
    'imt2020_single_element_pattern', 'imt2020_composite_pattern',
    ]


def _A_EH(phi, A_m, phi_3db):
    '''
    Antenna element's horizontal pattern according to `IMT.MODEL
    <https://www.itu.int/md/R15-TG5.1-C-0036>`_ document.

    Parameters
    ----------
    phi : np.ndarray, float
        Azimuth [deg]
    A_m : np.ndarray, float
        Front-to-back ratio (horizontal) [dimless]
    phi_3db : np.ndarray, float
        Horizontal 3-dB beam width of single element [deg]

    Returns
    -------
    A_EH : np.ndarray, float
        Antenna element's horizontal radiation pattern [dB]
    '''

    return -np.minimum(12 * (phi / phi_3db) ** 2, A_m)


def _A_EV(theta, SLA_nu, theta_3db):
    '''
    Antenna element's vertical pattern according to `IMT.MODEL
    <https://www.itu.int/md/R15-TG5.1-C-0036>`_ document.

    Parameters
    ----------
    theta : np.ndarray, float
        Elevation [deg]
    SLA_nu : np.ndarray, float
        Front-to-back ratio (vertical) [dimless]
    theta_3db : np.ndarray, float
        Vertical 3-dB beam width of single element [deg]

    Returns
    -------
    A_EV : np.ndarray, float
        Antenna element's vertical radiation pattern [dB]
    '''

    return -np.minimum(12 * ((theta - 90.) / theta_3db) ** 2, SLA_nu)


def _imt2020_single_element_pattern(
        azim, elev,
        G_Emax,
        A_m, SLA_nu,
        phi_3db, theta_3db
        ):

    phi = azim
    theta = 90. - elev
    return G_Emax - np.minimum(
        -_A_EH(phi, A_m, phi_3db) - _A_EV(theta, SLA_nu, theta_3db),
        A_m
        )


@utils.ranged_quantity_input(
    azim=(-180, 180, apu.deg),
    elev=(-90, 90, apu.deg),
    G_Emax=(None, None, cnv.dB),
    A_m=(0, None, cnv.dimless),
    SLA_nu=(0, None, cnv.dimless),
    phi_3db=(0, None, apu.deg),
    theta_3db=(0, None, apu.deg),
    strip_input_units=True, output_unit=cnv.dB
    )
def imt2020_single_element_pattern(
        azim, elev,
        G_Emax,
        A_m, SLA_nu,
        phi_3db, theta_3db
        ):
    '''
    Single antenna element's pattern according to `IMT.MODEL
    <https://www.itu.int/md/R15-TG5.1-C-0036>`_ document.

    Parameters
    ----------
    azim : `~astropy.units.Quantity`
        Azimuth [deg]
    elev : `~astropy.units.Quantity`
        Elevation [deg]
    G_Emax : `~astropy.units.Quantity`
        Single element maximum gain [dBi]
    A_m : `~astropy.units.Quantity`
        Front-to-back ratio (horizontal) [dimless]
    SLA_nu : `~astropy.units.Quantity`
        Front-to-back ratio (vertical) [dimless]
    phi_3db : `~astropy.units.Quantity`
        Horizontal 3dB beam width of single element [deg]
    theta_3db : `~astropy.units.Quantity`
        Vertical 3dB beam width of single element [deg]

    Returns
    -------
    A_E : `~astropy.units.Quantity`
        Single antenna element's pattern [dB]
    '''

    return _imt2020_single_element_pattern(
        azim, elev,
        G_Emax,
        A_m, SLA_nu,
        phi_3db, theta_3db
        )


def _imt2020_composite_pattern(
        azim, elev,
        azim_i, elev_i,
        G_Emax,
        A_m, SLA_nu,
        phi_3db, theta_3db,
        d_H, d_V,
        N_H, N_V,
        ):

    phi = azim
    theta = 90. - elev
    phi_i = azim_i
    theta_i = elev_i  # sic! (tilt angle in imt.model is elevation)

    A_E = _imt2020_single_element_pattern(
        azim, elev,
        G_Emax,
        A_m, SLA_nu,
        phi_3db, theta_3db
        )

    def nu(m, n):
        '''m, n zero-based (unlike in IMT.MODEL document, Table 4'''

        return np.exp(
            1j * 2 * np.pi * (
                n * d_V * np.cos(np.radians(theta)) +
                m * d_H * np.sin(np.radians(theta)) * np.sin(np.radians(phi))
                )
            )

    def w(m, n):
        '''m, n zero-based (unlike in IMT.MODEL document, Table 4'''

        return np.exp(
            1j * 2 * np.pi * (
                n * d_V * np.sin(np.radians(theta_i)) +
                m * d_H * np.cos(np.radians(theta_i)) *
                np.sin(np.radians(phi_i))
                )
            ) / np.sqrt(N_H * N_V)

    tmp = np.zeros(np.broadcast(phi, theta).shape, dtype=np.complex64)
    for m in range(N_H):
        for n in range(N_V):
            tmp += w(m, n) * nu(m, n)

    return A_E + 10 * np.log10(np.abs(tmp) ** 2)


@utils.ranged_quantity_input(
    azim=(-180, 180, apu.deg),
    elev=(-90, 90, apu.deg),
    azim_i=(-180, 180, apu.deg),
    elev_i=(-90, 90, apu.deg),
    G_Emax=(None, None, cnv.dB),
    A_m=(0, None, cnv.dimless),
    SLA_nu=(0, None, cnv.dimless),
    phi_3db=(0, None, apu.deg),
    theta_3db=(0, None, apu.deg),
    d_H=(0, None, cnv.dimless),
    d_V=(0, None, cnv.dimless),
    strip_input_units=True, output_unit=cnv.dB
    )
def imt2020_composite_pattern(
        azim, elev,
        azim_i, elev_i,
        G_Emax,
        A_m, SLA_nu,
        phi_3db, theta_3db,
        d_H, d_V,
        N_H, N_V,
        ):
    '''
    Composite (array) antenna pattern according to `IMT.MODEL
    <https://www.itu.int/md/R15-TG5.1-C-0036>`_ document.

    Parameters
    ----------
    azim, elev : `~astropy.units.Quantity`
        Azimuth/Elevation [deg]
    azim_i, elev_i : `~astropy.units.Quantity`
        Azimuthal/Elevational pointing of beam `i` [deg]
    G_Emax : `~astropy.units.Quantity`
        Single element maximum gain [dBi]
    A_m, SLA_nu : `~astropy.units.Quantity`
        Front-to-back ratio (horizontal/vertical) [dimless]
    phi_3db, theta_3db : `~astropy.units.Quantity`
        Horizontal/Vertical 3dB beam width of single element [deg]
    d_H, d_V : `~astropy.units.Quantity`
        Horizontal/Vertical separation of beams in units of wavelength
        [dimless]
    N_H, N_V : int
        Horizontal/Vertical number of single antenna elements

    Returns
    -------
    A_A : `~astropy.units.Quantity`
        Composite (array) antenna pattern of beam `i` [dB]
    '''

    return _imt2020_composite_pattern(
        azim, elev,
        azim_i, elev_i,
        G_Emax,
        A_m, SLA_nu,
        phi_3db, theta_3db,
        d_H, d_V,
        N_H, N_V,
        )


if __name__ == '__main__':
    print('This not a standalone python program! Use as module.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import (
    absolute_import, unicode_literals, division, print_function
    )

# from functools import partial, lru_cache
from astropy import units as apu
import numpy as np
from .. import conversions as cnv
from .. import utils


__all__ = [
    'fl_pattern',
    'fl_G_max_from_hpbw', 'fl_G_max_from_size', 'fl_hpbw_from_size',
    ]


def _fl_pattern_2_1(phi, diameter_over_wavelength, G_max):
    '''
    Antenna gain as a function of angular distance after `ITU-R Rec F.699
    <https://www.itu.int/rec/R-REC-F.699-7-200604-I/en>`_ 2.1.

    Valid for 1 to 70 GHz and cases where D / wavelength > 100
    '''

    phi = np.abs(phi)

    # the following are independent on phi, no need to compute after broadcast
    d_wlen = diameter_over_wavelength

    g1 = 2. + 15. * np.log10(d_wlen)  # gain of first side-lobe
    phi_m = 20. / d_wlen * np.sqrt(G_max - g1)
    phi_r = 15.85 * d_wlen ** -0.6

    # note automatic broadcasting should be possible
    (
        phi, d_wlen, G_max, g1, phi_m, phi_r,
        ) = np.broadcast_arrays(
        phi, d_wlen, G_max, g1, phi_m, phi_r,
        )
    gain = np.zeros(phi.shape, np.float64)

    # case 1:
    mask = (0 <= phi) & (phi < phi_m)
    gain[mask] = G_max[mask] - 2.5e-3 * (d_wlen[mask] * phi[mask]) ** 2

    # case 2:
    mask = (phi_m <= phi) & (phi < phi_r)
    gain[mask] = g1[mask]

    # case 3:
    mask = (phi_r <= phi) & (phi < 48.)
    gain[mask] = 32. - 25. * np.log10(phi[mask])

    # case 4:
    mask = (48. <= phi) & (phi <= 180.)
    gain[mask] = -10.

    return gain


def _fl_pattern_2_2(phi, diameter_over_wavelength, G_max):
    '''
    Antenna gain as a function of angular distance after `ITU-R Rec F.699
    <https://www.itu.int/rec/R-REC-F.699-7-200604-I/en>`_ 2.2.

    Valid for 1 to 70 GHz and cases where D / wavelength <= 100
    '''

    phi = np.abs(phi)

    # the following are independent on phi, no need to compute after broadcast
    d_wlen = diameter_over_wavelength

    g1 = 2. + 15. * np.log10(d_wlen)  # gain of first side-lobe
    phi_m = 20. / d_wlen * np.sqrt(G_max - g1)
    phi_r = 15.85 * d_wlen ** -0.6

    # note automatic broadcasting should be possible
    (
        phi, d_wlen, G_max, g1, phi_m, phi_r,
        ) = np.broadcast_arrays(
        phi, d_wlen, G_max, g1, phi_m, phi_r,
        )
    gain = np.zeros(phi.shape, np.float64)

    # case 1:
    mask = (0 <= phi) & (phi < phi_m)
    gain[mask] = G_max[mask] - 2.5e-3 * (d_wlen[mask] * phi[mask]) ** 2

    # case 2:
    mask = (phi_m <= phi) & (phi < phi_r)
    gain[mask] = g1[mask]

    # case 3:
    mask = (phi_r <= phi) & (phi < 48.)
    gain[mask] = 52 - 10 * np.log10(d_wlen[mask]) - 25 * np.log10(phi[mask])

    # case 4:
    mask = (48. <= phi) & (phi <= 180.)
    gain[mask] = -10. - 10 * np.log10(d_wlen[mask])

    return gain


def _fl_pattern_2_3(phi, diameter_over_wavelength, G_max):
    '''
    Antenna gain as a function of angular distance after `ITU-R Rec F.699
    <https://www.itu.int/rec/R-REC-F.699-7-200604-I/en>`_ 2.3.

    Valid for 0.1 to 1 GHz and cases where D / wavelength > 0.63
    '''

    phi = np.abs(phi)

    # the following are independent on phi, no need to compute after broadcast
    d_wlen = diameter_over_wavelength

    g1 = 2. + 15. * np.log10(d_wlen)  # gain of first side-lobe
    phi_m = 20. / d_wlen * np.sqrt(G_max - g1)
    phi_t = 100. / d_wlen
    phi_s = 144.5 * d_wlen ** -0.2

    # note automatic broadcasting should be possible
    (
        phi, d_wlen, G_max, g1, phi_m, phi_t, phi_s,
        ) = np.broadcast_arrays(
        phi, d_wlen, G_max, g1, phi_m, phi_t, phi_s,
        )
    gain = np.full(phi.shape, np.nan, np.float64)

    # case 1:
    mask = (0 <= phi) & (phi < phi_m)
    gain[mask] = G_max[mask] - 2.5e-3 * (d_wlen[mask] * phi[mask]) ** 2

    # case 2:
    mask = (phi_m <= phi) & (phi < phi_t)
    gain[mask] = g1[mask]

    # case 3:
    mask = (phi_t <= phi) & (phi < phi_s)
    gain[mask] = 52 - 10 * np.log10(d_wlen[mask]) - 25 * np.log10(phi[mask])

    # case 4:
    mask = (phi_s <= phi) & (phi <= 180.)
    gain[mask] = -2. - 5 * np.log10(d_wlen[mask])

    return gain


def _fl_pattern(phi, diameter, wavelength, G_max):

    d_wlen = diameter / wavelength

    phi, d_wlen, G_max = np.broadcast_arrays(phi, d_wlen, G_max)
    gain = np.full(phi.shape, np.nan, np.float64)

    tmp_mask = (0.00428 < wavelength) & (wavelength < 0.29979)  # 1...70 GHz
    mask_2_1 = (d_wlen > 100) & tmp_mask
    mask_2_2 = (d_wlen <= 100) & tmp_mask

    tmp_mask = (0.29979 <= wavelength) & (wavelength < 2.99792)  # 0.1...1 GHz
    mask_2_3 = (d_wlen > 0.63) & tmp_mask

    for mask, func in [
            (mask_2_1, _fl_pattern_2_1),
            (mask_2_2, _fl_pattern_2_2),
            (mask_2_3, _fl_pattern_2_3),
            ]:

        gain[mask] = func(phi[mask], d_wlen[mask], G_max[mask])

    return gain


@utils.ranged_quantity_input(
    phi=(-180, 180, apu.deg),
    diameter=(0.1, 1000., apu.m),
    wavelength=(0.001, 2, apu.m),
    G_max=(None, None, cnv.dBi),
    strip_input_units=True, output_unit=cnv.dBi
    )
def fl_pattern(
        phi, diameter, wavelength, G_max
        ):
    '''
    Antenna gain as a function of angular distance after `ITU-R Rec F.699
    <https://www.itu.int/rec/R-REC-F.699-7-200604-I/en>`_.

    Parameters
    ----------
    phi : `~astropy.units.Quantity`
        Angular distance[deg]
    diameter : `~astropy.units.Quantity`
        Antenna diameter [m]
    wavelength : `~astropy.units.Quantity`
        Observing wavelength [m]
    G_max : `~astropy.units.Quantity`
        Antenna maximum gain [dBi]

    Returns
    -------
    gain : `~astropy.units.Quantity`
        Antenna gain [dBi]

    Notes
    -----
    - If `G_max` is unknown, there is a helper function
      `pycraf.antenna.G_max_from_HPBW` that can be used to estimate its
      value.
    - See `ITU-R Rec F.699
      <https://www.itu.int/rec/R-REC-F.699-7-200604-I/en>`_ for further
      explanation and applicability of this model.
    '''

    return _fl_pattern(phi, diameter, wavelength, G_max)


def _fl_hpbw_from_size(diameter, wavelength):

    return 70. * wavelength / diameter


@utils.ranged_quantity_input(
    diameter=(0.1, 1000., apu.m),
    wavelength=(0.001, 2, apu.m),
    strip_input_units=True, output_unit=apu.deg
    )
def fl_hpbw_from_size(diameter, wavelength):
    '''
    Estimate HPBW from antenna diameter after `ITU-R Rec F.699
    <https://www.itu.int/rec/R-REC-F.699-7-200604-I/en>`_.

    Parameters
    ----------
    diameter : `~astropy.units.Quantity`
        Antenna diameter [m]
    wavelength : `~astropy.units.Quantity`
        Observing wavelength [m]

    Returns
    -------
    hpbw : `~astropy.units.Quantity`
        Antenna HPBW (3-dB point) [deg]

    Notes
    -----
    See `ITU-R Rec F.699
    <https://www.itu.int/rec/R-REC-F.699-7-200604-I/en>`_ for further
    explanation and applicability of this model.
    '''

    return _fl_hpbw_from_size(diameter, wavelength)


def _fl_G_max_from_size(diameter, wavelength):

    return 20. * np.log10(diameter / wavelength) + 7.7


@utils.ranged_quantity_input(
    diameter=(0.1, 1000., apu.m),
    wavelength=(0.001, 2, apu.m),
    strip_input_units=True, output_unit=cnv.dBi
    )
def fl_G_max_from_size(diameter, wavelength):
    '''
    Estimate maximum gain from antenna diameter after `ITU-R Rec F.699
    <https://www.itu.int/rec/R-REC-F.699-7-200604-I/en>`_.

    Parameters
    ----------
    diameter : `~astropy.units.Quantity`
        Antenna diameter [m]
    wavelength : `~astropy.units.Quantity`
        Observing wavelength [m]

    Returns
    -------
    G_max : `~astropy.units.Quantity`
        Antenna maximum gain [dBi]

    Notes
    -----
    See `ITU-R Rec F.699
    <https://www.itu.int/rec/R-REC-F.699-7-200604-I/en>`_ for further
    explanation and applicability of this model.
    '''

    return _fl_G_max_from_size(diameter, wavelength)


def _fl_G_max_from_hpbw(hpbw):

    return 44.5 - 20 * np.log10(hpbw)


@utils.ranged_quantity_input(
    hpbw=(1.e-3, 90., apu.deg),
    strip_input_units=True, output_unit=cnv.dBi
    )
def fl_G_max_from_hpbw(hpbw):
    '''
    Estimate maximum gain from antenna hpbw after `ITU-R Rec F.699
    <https://www.itu.int/rec/R-REC-F.699-7-200604-I/en>`_.

    Parameters
    ----------
    hpbw : `~astropy.units.Quantity`
        Antenna HPBW (3-dB point) [deg]

    Returns
    -------
    G_max : `~astropy.units.Quantity`
        Antenna maximum gain [dBi]

    Notes
    -----
    See `ITU-R Rec F.699
    <https://www.itu.int/rec/R-REC-F.699-7-200604-I/en>`_ for further
    explanation and applicability of this model.
    '''

    return _fl_G_max_from_hpbw(hpbw)


if __name__ == '__main__':
    print('This not a standalone python program! Use as module.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
This sub-package provides various (simplified) antenna patterns for use in
compatibility studies. At the moment, the following are implemented:

- IMT phased array antenna patterns for base station and mobile devices to
  be used for AI 1.13 IMT2020 studies. These are defined in the so-called
  IMT.MODEL document (`ITU-R TG 5/1 document 5-1/36 <https://www.itu.int/md/R15-TG5.1-C-0036>`_).
- A pattern that can be used for radio telescopes (`ITU-R Rec. RA.1631-0
  <https://www.itu.int/rec/R-REC-RA.1631-0-200305-I/en>`_).
- Pattern for fixed wireless systems ("fixed-links"), as defined in
  `ITU-R Rec. F.699-7 <https://www.itu.int/rec/R-REC-F.699-7-200604-I/en>`_.
'''

from .imt import *
from .ras import *
from .fixedlink import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import (
    absolute_import, unicode_literals, division, print_function
    )

from astropy import units as apu
import numpy as np
from .. import utils
from .cygeodesics import inverse_cython, direct_cython


__all__ = ['geoid_inverse', 'geoid_direct']


@utils.ranged_quantity_input(
    lon1=(-np.pi, np.pi, apu.rad),
    lat1=(-np.pi / 2, np.pi / 2, apu.rad),
    lon2=(-np.pi, np.pi, apu.rad),
    lat2=(-np.pi / 2, np.pi / 2, apu.rad),
    strip_input_units=True,
    output_unit=(apu.m, apu.rad, apu.rad)
    )
def geoid_inverse(
        lon1, lat1,
        lon2, lat2,
        eps=1.e-12,  # corresponds to approximately 0.06mm
        maxiter=50,
        ):
    '''
    Solve inverse Geodesics problem using Vincenty's formulae.

    Using an iterative approach, the distance and relative bearings between
    two points (P1, and P2) on the Geoid (Earth ellipsoid) are determined;
    see also `Wikipedia <https://en.wikipedia.org/wiki/Vincenty's_formulae>`_.

    Parameters
    ----------
    lon1 : `~astropy.units.Quantity`
        Geographic longitude of P1 [rad]
    lat1 : `~astropy.units.Quantity`
        Geographic latitude of P1 [rad]
    lon2 : `~astropy.units.Quantity`
        Geographic longitude of P2 [rad]
    lat2 : `~astropy.units.Quantity`
        Geographic latitude of P2 [rad]
    eps : float, optional
        Accuracy of calculation (default: 1.e-12)
    maxiter : int, optional
        Maximum number of iterations to perform (default: 50)

    Returns
    -------
    distance : `~astropy.units.Quantity`
        Distance between P1 and P2 [m]
    bearing1 : `~astropy.units.Quantity`
        Start bearing [rad]
    bearing2 : `~astropy.units.Quantity`
        Back-bearing [rad]

    Notes
    -----
    The iteration will stop if either the desired accuracy (`eps`) is reached
    or the number of iterations exceeds `maxiter`.
    '''

    return inverse_cython(lon1, lat1, lon2, lat2, eps=eps, maxiter=maxiter)


@utils.ranged_quantity_input(
    lon1=(-np.pi, np.pi, apu.rad),
    lat1=(-np.pi / 2, np.pi / 2, apu.rad),
    bearing1=(-np.pi, np.pi, apu.rad),
    dist=(0.1, None, apu.m),
    strip_input_units=True,
    output_unit=(apu.rad, apu.rad, apu.rad)
    )
def geoid_direct(
        lon1, lat1,
        bearing1, dist,
        eps=1.e-12,  # corresponds to approximately 0.06mm
        maxiter=50,
        ):
    '''
    Solve direct Geodesics problem using Vincenty's formulae.

    From starting point P1, given a start bearing, find
    point P2 located at a certain distance from P1 on the Geoid (Earth
    ellipsoid). As for the inverse problem, an iterative approach is used;
    see also `Wikipedia <https://en.wikipedia.org/wiki/Vincenty's_formulae>`_.

    Parameters
    ----------
    lon1 : `~astropy.units.Quantity`
        Geographic longitude of P1 [rad]
    lat1 : `~astropy.units.Quantity`
        Geographic latitude of P1 [rad]
    bearing1 : `~astropy.units.Quantity`
        Start bearing [rad]
    distance : `~astropy.units.Quantity`
        Distance between P1 and P2 [m]
    eps : float, optional
        Accuracy of calculation (default: 1.e-12)
    maxiter : int, optional
        Maximum number of iterations to perform (default: 50)

    Returns
    -------
    lon2 : `~astropy.units.Quantity`
        Geographic longitude of P2 [rad]
    lat2 : `~astropy.units.Quantity`
        Geographic latitude of P2 [rad]
    bearing2 : `~astropy.units.Quantity`
        Back-bearing [rad]

    Notes
    -----
    The iteration will stop if either the desired accuracy (`eps`) is reached
    or the number of iterations exceeds `maxiter`.
    '''

    return direct_cython(
        lon1, lat1, bearing1, dist, eps=eps, maxiter=maxiter, wrap=True
        )


if __name__ == '__main__':
    print('This not a standalone python program! Use as module.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function
# from __future__ import unicode_literals

import pytest
import numpy as np
from numpy.testing import assert_equal, assert_allclose
from astropy.tests.helper import assert_quantity_allclose
from astropy import units as apu
from astropy.units import Quantity
from ... import conversions as cnv
from ... import protection as prot
from ...utils import check_astro_quantities
# from astropy.utils.misc import NumpyRNGContext


TOL_KWARGS = {'atol': 0., 'rtol': 1.e-6}


@pytest.mark.skip(reason='failing on AppVeyor and Travis for unknown reason')
def test_cispr_limits():

    args_list = [
        (0, None, apu.Hz),
        ]
    kwargs_list = [
        ('detector_dist', 0.001, None, apu.m),
        ]
    check_astro_quantities(prot.cispr11_limits, args_list, kwargs_list)
    check_astro_quantities(prot.cispr22_limits, args_list, kwargs_list)

    res = prot.cispr11_limits(1 * apu.Hz)
    assert isinstance(res, tuple)
    assert res[0].unit == cnv.dB_uV_m
    res = prot.cispr22_limits(1 * apu.Hz)
    assert isinstance(res, tuple)
    assert res[0].unit == cnv.dB_uV_m

    freqs = Quantity(np.linspace(10, 900, 7), apu.MHz)

    assert_quantity_allclose(
        prot.cispr11_limits(
            freqs,
            detector_type='RMS',
            detector_dist=30. * apu.m
            )[0],
        np.array([
            24.5, 24.5, 31.5, 31.5, 31.5, 31.5, 31.5
            ]) * cnv.dB_uV_m,
        )

    assert_quantity_allclose(
        prot.cispr11_limits(
            freqs,
            detector_type='RMS',
            detector_dist=300. * apu.m
            )[0],
        np.array([
            4.5, 4.5, 11.5, 11.5, 11.5, 11.5, 11.5
            ]) * cnv.dB_uV_m,
        )

    assert_quantity_allclose(
        prot.cispr11_limits(
            freqs,
            detector_type='QP',
            detector_dist=30. * apu.m
            )[0],
        np.array([
            30., 30., 37., 37., 37., 37., 37.
            ]) * cnv.dB_uV_m,
        )

    freqs = Quantity(np.linspace(10, 3100, 7), apu.MHz)

    assert_quantity_allclose(
        prot.cispr22_limits(
            freqs,
            detector_type='RMS',
            detector_dist=30. * apu.m
            )[0],
        np.array([
            34.5, 41.5, 50.5, 50.5, 50.5, 50.5, 54.5
            ]) * cnv.dB_uV_m,
        )

    assert_quantity_allclose(
        prot.cispr22_limits(
            freqs,
            detector_type='RMS',
            detector_dist=300. * apu.m
            )[0],
        np.array([
            14.5, 21.5, 30.5, 30.5, 30.5, 30.5, 34.5
            ]) * cnv.dB_uV_m,
        )

    assert_quantity_allclose(
        prot.cispr22_limits(
            freqs,
            detector_type='QP',
            detector_dist=30. * apu.m
            )[0],
        np.array([
            40., 47., 56., 56., 56., 56., 60.
            ]) * cnv.dB_uV_m,
        )


CONT_LIMS_VALUES = [
    5006.0, 5.1822630178e-21, 7.4925620118e-18, 1594.88178292, 28.200091414,
    ]
CONT_LIMS_DB_VALUES = [
    5006.0, -202.854805488, -171.25369654, -227.972715025, 29.0050103227,
    ]
SPEC_LIMS_VALUES = [
    0.7071067812, 3.03742736e-21, 2.186381732e-15, 4225.3020266, 12.522083029,
    ]
SPEC_LIMS_DB_VALUES = [
    0.707106781187, -205.174940994, -146.6027401, -223.7414224, 21.9535315816,
    ]

COL_NAMES = [
    'frequency',
    'bandwidth',
    'T_A',
    'T_rx',
    'T_rms',
    'P_rms_nu',
    'Plim',
    'Slim',
    'Slim_nu',
    'Efield',
    'Efield_norm',
    ]

COL_UNITS = [
    apu.MHz,
    apu.MHz,
    apu.K,
    apu.K,
    apu.mK,
    apu.W / apu.Hz,
    apu.W,
    apu.W / apu.m ** 2,
    apu.Jy,
    apu.uV / apu.m,
    apu.uV / apu.m,
    ]

COL_UNITS_DB = [
    apu.MHz,
    apu.MHz,
    apu.K,
    apu.K,
    apu.mK,
    cnv.dB_W_Hz,
    cnv.dB_W,
    cnv.dB_W_m2,
    cnv.dB_W_m2_Hz,
    cnv.dB_uV_m,
    cnv.dB_uV_m,
    ]


def test_ra769_limits():

    with pytest.raises(AssertionError):
        prot.ra769_limits(mode='FOO')

    with pytest.raises(AssertionError):
        prot.ra769_limits(scale='FOO')

    cont_lims = prot.ra769_limits(mode='continuum', scale='linear')
    cont_lims_dB = prot.ra769_limits(mode='continuum', scale='dB')
    spec_lims = prot.ra769_limits(mode='spectroscopy', scale='linear')
    spec_lims_dB = prot.ra769_limits(mode='spectroscopy', scale='dB')

    # test some "random" elements in all tables (values and units)
    # prefer derived quantities
    for idx, (row, col) in enumerate([
            (0, 'T_rms'),
            (5, 'Plim'),
            (10, 'Slim'),
            (15, 'Slim_nu'),
            (20, 'Efield'),
            ]):
        assert_allclose(cont_lims[row][col], CONT_LIMS_VALUES[idx], **TOL_KWARGS)
        assert_allclose(cont_lims_dB[row][col], CONT_LIMS_DB_VALUES[idx])

    for idx, (row, col) in enumerate([
            (0, 'T_rms'),
            (3, 'Plim'),
            (6, 'Slim'),
            (9, 'Slim_nu'),
            (12, 'Efield'),
            ]):
        assert_allclose(spec_lims[row][col], SPEC_LIMS_VALUES[idx], **TOL_KWARGS)
        assert_allclose(spec_lims_dB[row][col], SPEC_LIMS_DB_VALUES[idx])

    for colname, colunit in zip(COL_NAMES, COL_UNITS):
        assert cont_lims.columns[colname].unit == colunit
        assert spec_lims.columns[colname].unit == colunit

    for colname, colunit in zip(COL_NAMES, COL_UNITS_DB):
        assert cont_lims_dB.columns[colname].unit == colunit
        assert spec_lims_dB.columns[colname].unit == colunit



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Contains functions to convert power flux densities, field strengths,
transmitted and received powers at certain distances and frequencies into each
other. Furthermore, some convenience decibel-Scales are defined.

The following Decibel units are available:


================ ============== ======================================
      Unit             Alias         Definition
================ ============== ======================================
|dimless|        dimless        |u|.dimensionless_unscaled
|dB| = |dBi|     dB = dBi       |u|.dB(dimless)
|dBW|            dB_W           |u|.dB(|u|.W)
|dBW_Hz|         dB_W_Hz        |u|.dB(|u|.W / |u|.Hz)
|dBW_m2|         dB_W_m2        |u|.dB(|u|.W / |u|.m ** 2)
|dBW_m2_Hz|      dB_W_m2_Hz     |u|.dB(|u|.W / |u|.m ** 2 / |u|.Hz)
|dBW_Jy_Hz|      dB_Jy_Hz       |u|.dB(|u|.Jy * |u|.Hz)
|dBm|            dBm = dB_mW    |u|.dB(|u|.mW)
|dBm_MHz|        dBm_MHz        |u|.dB(|u|.mW / |u|.MHz)
|dB_uV_m|        dB_uV_m        |u|.dB(|u|.uV ** 2 / |u|.m ** 2)
|dB_1_m|         dB_1_m         |u|.dB(1 / |u|.m)
================ ============== ======================================

u = `astropy.units <http://docs.astropy.org/en/stable/units/index.html>`_

.. |u| replace:: u
.. |dimless| replace:: :math:`1`
.. |dB| replace:: :math:`\mathrm{dB}`
.. |dBi| replace:: :math:`\mathrm{dBi}`
.. |dBW| replace:: :math:`\mathrm{dB}[\mathrm{W}]`
.. |dBW_Hz| replace:: :math:`\mathrm{dB}[\mathrm{W} / \mathrm{Hz}]`
.. |dBW_m2| replace:: :math:`\mathrm{dB}[\mathrm{W} / \mathrm{m}^2]`
.. |dBW_m2_Hz| replace:: :math:`\mathrm{dB}[\mathrm{W} / \mathrm{m}^2 / \mathrm{Hz}]`
.. |dBW_Jy_Hz| replace:: :math:`\mathrm{dB}[\mathrm{Jy} \cdot \mathrm{Hz}]`
.. |dBm| replace:: :math:`\mathrm{dB}[\mathrm{mW}]`
.. |dBm_MHz| replace:: :math:`\mathrm{dB}[\mathrm{mW} / \mathrm{MHz}]`
.. |dB_uV_m| replace:: :math:`\mathrm{dB}[\mu\mathrm{V}^2 / \mathrm{m}^2]`
.. |dB_1_m| replace:: :math:`\mathrm{dB}[1 / \mathrm{m}]`
'''

from .conversions import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function
# from __future__ import unicode_literals

import os
import pytest
from functools import partial
import numpy as np
from numpy.testing import assert_equal, assert_allclose
from astropy.tests.helper import assert_quantity_allclose
from astropy import units as apu
from astropy.units import Quantity
from ... import conversions as cnv
from ... import pathprof
from ...utils import check_astro_quantities, get_pkg_data_filename
from astropy.utils.misc import NumpyRNGContext
import json
from itertools import product
import h5py


TOL_KWARGS = {'atol': 1.e-4, 'rtol': 1.e-4}


class TestPropagation:

    def setup(self):

        # TODO: add further test cases

        self.lon_t, self.lat_t = 6.5 * apu.deg, 50.5 * apu.deg
        self.lon_r, self.lat_r = 6.6 * apu.deg, 50.75 * apu.deg
        self.hprof_step = 100 * apu.m

        self.omega = 0. * apu.percent
        self.temperature = (273.15 + 15.) * apu.K  # as in Excel sheet
        self.pressure = 1013. * apu.hPa

        self.cases = list(
            # freq, (h_tg, h_rg), time_percent, version, (G_t, G_r)
            product(
                [0.1, 1., 10.],
                [(50, 50), (200, 200)],
                [2, 10, 50],
                [14, 16],
                [(0, 0), (20, 30)],
            ))

        self.pprop_template = (
            'cases/pprop_{:.2f}ghz_{:.2f}m_{:.2f}m_{:.2f}percent_v{:d}.json'
            )
        self.loss_template = (
            'cases/loss_{:.2f}ghz_{:.2f}m_{:.2f}m_{:.2f}percent_'
            '{:.2f}db_{:.2f}db_v{:d}.json'
            )
        self.fastmap_template = (
            'fastmap/attens_{:.2f}ghz_{:.2f}m_{:.2f}m_{:.2f}percent_v{:d}.hdf5'
            )
        self.pprops = []
        for case in self.cases:

            freq, (h_tg, h_rg), time_percent, version, (G_t, G_r) = case

            pprop = pathprof.PathProp(
                freq * apu.GHz,
                self.temperature, self.pressure,
                self.lon_t, self.lat_t,
                self.lon_r, self.lat_r,
                h_tg * apu.m, h_rg * apu.m,
                self.hprof_step,
                time_percent * apu.percent,
                version=version,
                )
            self.pprops.append(pprop)

            # Warning: if uncommenting, the test cases will be overwritten
            # do this only, if you need to update the json files
            # (make sure, that results are correct!)
            # pprop_name = self.pprop_template.format(
            #     freq, h_tg, h_rg, time_percent, version
            #     )
            # json.dump(pprop._pp, open(pprop_name, 'w'))

            los_loss = pathprof.loss_freespace(pprop)
            trop_loss = pathprof.loss_troposcatter(
                pprop, G_t * cnv.dB, G_r * cnv.dB,
                )
            duct_loss = pathprof.loss_ducting(pprop)
            diff_loss = pathprof.loss_diffraction(pprop)

            tot_loss = pathprof.loss_complete(
                pprop, G_t * cnv.dB, G_r * cnv.dB,
                )

            losses = {}
            losses['L_bfsg'] = los_loss[0].to(cnv.dB).value
            losses['E_sp'] = los_loss[1].to(cnv.dB).value
            losses['E_sbeta'] = los_loss[2].to(cnv.dB).value

            losses['L_bs'] = trop_loss.to(cnv.dB).value
            losses['L_ba'] = duct_loss.to(cnv.dB).value

            losses['L_d_50'] = diff_loss[0].to(cnv.dB).value
            losses['L_dp'] = diff_loss[1].to(cnv.dB).value
            losses['L_bd_50'] = diff_loss[2].to(cnv.dB).value
            losses['L_bd'] = diff_loss[3].to(cnv.dB).value
            losses['L_min_b0p'] = diff_loss[4].to(cnv.dB).value

            losses['L_bfsg_t'] = tot_loss[0].to(cnv.dB).value
            losses['L_bd_t'] = tot_loss[1].to(cnv.dB).value
            losses['L_bs_t'] = tot_loss[2].to(cnv.dB).value
            losses['L_ba_t'] = tot_loss[3].to(cnv.dB).value
            losses['L_b_t'] = tot_loss[4].to(cnv.dB).value
            losses['L_b_corr_t'] = tot_loss[5].to(cnv.dB).value
            losses['L_t'] = tot_loss[6].to(cnv.dB).value

            # Warning: if uncommenting, the test cases will be overwritten
            # do this only, if you need to update the json files
            # (make sure, that results are correct!)
            # loss_name = self.loss_template.format(
            #     freq, h_tg, h_rg, time_percent, version, G_t, G_r
            #     )
            # json.dump(losses, open(loss_name, 'w'))

    def teardown(self):

        pass

    def test_pathprop(self):

        for freq, (h_tg, h_rg), time_percent, version, _ in self.cases:

            pprop = pathprof.PathProp(
                freq * apu.GHz,
                self.temperature, self.pressure,
                self.lon_t, self.lat_t,
                self.lon_r, self.lat_r,
                h_tg * apu.m, h_rg * apu.m,
                self.hprof_step,
                time_percent * apu.percent,
                version=version,
                )

            pprop_name = get_pkg_data_filename(
                self.pprop_template.format(
                    freq, h_tg, h_rg, time_percent, version
                    ))
            with open(pprop_name, 'r') as f:
                pprop_true = json.load(f)

            for k in pprop._pp:
                assert_quantity_allclose(pprop._pp[k], pprop_true[k])

    def test_freespace_loss(self):

        for case, pprop in zip(self.cases, self.pprops):

            freq, (h_tg, h_rg), time_percent, version, (G_t, G_r) = case
            los_loss = pathprof.loss_freespace(pprop)
            losses = {}
            losses['L_bfsg'] = los_loss[0].to(cnv.dB).value
            losses['E_sp'] = los_loss[1].to(cnv.dB).value
            losses['E_sbeta'] = los_loss[2].to(cnv.dB).value

            loss_name = get_pkg_data_filename(
                self.loss_template.format(
                    freq, h_tg, h_rg, time_percent, version, G_t, G_r
                    ))
            with open(loss_name, 'r') as f:
                loss_true = json.load(f)

            for k in losses:
                assert_quantity_allclose(losses[k], loss_true[k])

    def test_troposcatter_loss(self):

        for case, pprop in zip(self.cases, self.pprops):

            freq, (h_tg, h_rg), time_percent, version, (G_t, G_r) = case
            tropo_loss = pathprof.loss_troposcatter(
                pprop, G_t * cnv.dB, G_r * cnv.dB
                )
            losses = {}
            losses['L_bs'] = tropo_loss.to(cnv.dB).value

            loss_name = get_pkg_data_filename(
                self.loss_template.format(
                    freq, h_tg, h_rg, time_percent, version, G_t, G_r
                ))
            with open(loss_name, 'r') as f:
                loss_true = json.load(f)

            for k in losses:
                assert_quantity_allclose(losses[k], loss_true[k])

    def test_ducting_loss(self):

        for case, pprop in zip(self.cases, self.pprops):

            freq, (h_tg, h_rg), time_percent, version, (G_t, G_r) = case
            duct_loss = pathprof.loss_ducting(pprop)
            losses = {}
            losses['L_ba'] = duct_loss.to(cnv.dB).value

            loss_name = get_pkg_data_filename(
                self.loss_template.format(
                    freq, h_tg, h_rg, time_percent, version, G_t, G_r
                ))
            with open(loss_name, 'r') as f:
                loss_true = json.load(f)

            for k in losses:
                assert_quantity_allclose(losses[k], loss_true[k])

    def test_diffraction_loss(self):

        for case, pprop in zip(self.cases, self.pprops):

            freq, (h_tg, h_rg), time_percent, version, (G_t, G_r) = case
            diff_loss = pathprof.loss_diffraction(pprop)
            losses = {}
            losses['L_d_50'] = diff_loss[0].to(cnv.dB).value
            losses['L_dp'] = diff_loss[1].to(cnv.dB).value
            losses['L_bd_50'] = diff_loss[2].to(cnv.dB).value
            losses['L_bd'] = diff_loss[3].to(cnv.dB).value
            losses['L_min_b0p'] = diff_loss[4].to(cnv.dB).value

            loss_name = get_pkg_data_filename(
                self.loss_template.format(
                    freq, h_tg, h_rg, time_percent, version, G_t, G_r
                ))
            with open(loss_name, 'r') as f:
                loss_true = json.load(f)

            for k in losses:
                assert_quantity_allclose(losses[k], loss_true[k])

    def test_complete_loss(self):

        for case, pprop in zip(self.cases, self.pprops):

            freq, (h_tg, h_rg), time_percent, version, (G_t, G_r) = case
            tot_loss = pathprof.loss_complete(
                pprop, G_t * cnv.dB, G_r * cnv.dB
                )
            losses = {}
            losses['L_bfsg_t'] = tot_loss[0].to(cnv.dB).value
            losses['L_bd_t'] = tot_loss[1].to(cnv.dB).value
            losses['L_bs_t'] = tot_loss[2].to(cnv.dB).value
            losses['L_ba_t'] = tot_loss[3].to(cnv.dB).value
            losses['L_b_t'] = tot_loss[4].to(cnv.dB).value
            losses['L_b_corr_t'] = tot_loss[5].to(cnv.dB).value
            losses['L_t'] = tot_loss[6].to(cnv.dB).value

            loss_name = get_pkg_data_filename(
                self.loss_template.format(
                    freq, h_tg, h_rg, time_percent, version, G_t, G_r
                ))
            with open(loss_name, 'r') as f:
                loss_true = json.load(f)

            for k in losses:
                assert_quantity_allclose(losses[k], loss_true[k])

    def test_height_profile_data(self, tmpdir_factory):

        hprof_data_cache = pathprof.height_profile_data(
            6.5 * apu.deg, 50.5 * apu.deg,
            900 * apu.arcsec, 900 * apu.arcsec,
            map_resolution=30 * apu.arcsec,
            )

        # also testing reading/writing from hdf5
        tdir = tmpdir_factory.mktemp('hdata')
        tfile = str(tdir.join('hprof.hdf5'))
        print('writing temporary files to', tdir)

        # saving
        with h5py.File(tfile, 'w') as h5f:
            for k, v in hprof_data_cache.items():
                h5f[k] = v

        hprof_data_cache_disk = h5py.File(tfile, 'r')

        for k in hprof_data_cache:
            assert_quantity_allclose(
                # Note conversion to some ndarray type necessary, as h5py
                # returns <HDF5 dataset> types
                np.squeeze(hprof_data_cache[k]),
                np.squeeze(hprof_data_cache_disk[k])
                )

        # also test versus true results
        tfile = get_pkg_data_filename('fastmap/hprof.hdf5')
        hprof_data_cache_true = h5py.File(tfile, 'r')

        for k in hprof_data_cache:
            assert_quantity_allclose(
                # Note conversion to some ndarray type necessary, as h5py
                # returns <HDF5 dataset> types
                np.squeeze(hprof_data_cache[k]),
                np.squeeze(hprof_data_cache_true[k]),
                atol=1.e-6,
                )

    def test_fast_atten_map(self, tmpdir_factory):

        tfile = get_pkg_data_filename('fastmap/hprof.hdf5')
        hprof_data_cache = h5py.File(tfile, 'r')

        for case in self.cases:

            freq, (h_tg, h_rg), time_percent, version, (G_t, G_r) = case

            atten_map, eps_pt_map, eps_pr_map = pathprof.atten_map_fast(
                freq * apu.GHz,
                self.temperature,
                self.pressure,
                h_tg * apu.m, h_rg * apu.m,
                time_percent * apu.percent,
                hprof_data_cache,  # dict_like
                version=version,
                )

            atten_map = atten_map.to(cnv.dB).value
            eps_pt_map = eps_pt_map.to(apu.deg).value
            eps_pr_map = eps_pr_map.to(apu.deg).value

            tfile = get_pkg_data_filename(
                self.fastmap_template.format(
                    freq, h_tg, h_rg, time_percent, version
                    ))

            # Warning: if uncommenting, the test cases will be overwritten
            # do this only, if you need to update the json files
            # (make sure, that results are correct!)
            # with h5py.File(tfile, 'w') as h5f:
            #     h5f['atten_map'] = atten_map
            #     h5f['eps_pt_map'] = eps_pt_map
            #     h5f['eps_pr_map'] = eps_pr_map

            print(tfile)
            h5f = h5py.File(tfile, 'r')

            # Note conversion to some ndarray type necessary, as h5py
            # returns <HDF5 dataset> types
            tol_kwargs = {'atol': 1.e-6, 'rtol': 1.e-6}
            # atten_map[0, 0, 0] = 10
            # for some super-strange reason, index 9, 13 is completely off
            # on travis and appveyor (only diffraction)
            # as it is only one pixel, we ignore it here for now
            h5_atten_map = np.squeeze(h5f['atten_map'])
            h5_atten_map[:, 9, 13] = atten_map[:, 9, 13]

            idx = np.where(np.abs(h5_atten_map - atten_map) > 1.e-6)
            for i, y, x in zip(*idx):
                print(i, y, x, h5_atten_map[i, y, x], atten_map[i, y, x])
            assert_allclose(
                h5_atten_map, atten_map, **tol_kwargs
                )
            assert_allclose(h5f['eps_pt_map'], eps_pt_map)
            assert_allclose(h5f['eps_pr_map'], eps_pr_map)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
Allow cythonizing of our pyx files and provide custom compiler options.
'''

import os
from setuptools.extension import Extension
import platform
# Note: importing numpy from here won't work, see:
# http://docs.astropy.org/en/stable/development/ccython.html#using-numpy-c-headers
# import numpy as np
# 'include_dirs': [np.get_include()], --> 'include_dirs': 'numpy'

PYXDIR = os.path.relpath(os.path.dirname(__file__))


def get_extensions():

    comp_args = {
        'extra_compile_args': ['-fopenmp', '-O3'],
        'extra_link_args': ['-fopenmp'],
        'libraries': ['m'],
        'include_dirs': ['numpy'],
        }

    if platform.system().lower() == 'windows':
        comp_args = {
            'extra_compile_args': ['/openmp'],
            'include_dirs': ['numpy'],
            }
    elif 'darwin' in platform.system().lower():
        os.environ['CC'] = 'gcc-6'
        os.environ['LD'] = 'gcc-6'
        comp_args = {
            'extra_compile_args': [
                '-fopenmp', '-O3', '-mmacosx-version-min=10.7'
                ],
            'extra_link_args': ['-fopenmp'],
            'libraries': ['m'],
            'include_dirs': ['numpy'],
            }

    ext_module_pathprof_cyprop = Extension(
        name='pycraf.pathprof.cyprop',
        # ['pycraf/pathprof/cyprop.pyx'],
        sources=[os.path.join(PYXDIR, 'cyprop.pyx')],
        **comp_args
        )

    ext_module_pathprof_geodesics = Extension(
        name='pycraf.pathprof.cygeodesics',
        # ['pycraf/pathprof/geodesics.pyx'],
        sources=[os.path.join(PYXDIR, 'cygeodesics.pyx')],
        **comp_args
        )

    print('get_extensions', ext_module_pathprof_cyprop)
    return [ext_module_pathprof_cyprop, ext_module_pathprof_geodesics]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
This subpackage contains a couple of functions to convert coordinates
between typical GIS frames, such as GPS (WGS84), UTM, and ETRS89. The
underlying transformations are based on the `pyproj
<https://pypi.python.org/pypi/pyproj>`_ package, which itself is a wrapper
around the `proj.4 <http://proj4.org/>`_ software.
'''


from .geospatial import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import (
    absolute_import, unicode_literals, division, print_function
    )

import os
import numpy as np
from astropy import units as apu
# from astropy.units import Quantity, UnitsError
import astropy.constants as con
from astropy.table import QTable, Table
from .. import conversions as cnv


__all__ = ['ra769_limits']


def ra769_limits(mode='continuum', scale='dB'):
    '''
    Limits for spectral line and continuum observations according to
    `ITU-R Rec RA.769 <https://www.itu.int/rec/R-REC-RA.769-2-200305-I/en>`_.


    Parameters
    ----------
    mode : str, optional
        Observing mode: 'continuum', 'spectroscopy' (default: 'continuum')
    scale : str, optional
        Default scale to use: 'linear', 'dB' (default: 'linear')

    Returns
    -------
    ra769_limits : `~astropy.table.Table`
        A table with the following columns:

        - "frequency" [MHz]
        - "bandwidth" [MHz]
        - "T_A" [K]
        - "T_rx" [K]
        - "T_rms" [K]
        - "P_rms_nu" [W/Hz, dB(W/Hz)]
        - "Plim" [W, dB(W)]
        - "Plim_nu" [W/Hz, dB(W/Hz)]
        - "Slim" [W/m^2, dB(W/m^2)]
        - "Slim_nu" [Jy, dB(W/m^2/Hz)]
        - "Efield" [uV/m, db(uV^2/m^2)]
        - "Efield_norm", normalized to 1 MHz bandwidth [uV/m, db(uV^2/m^2)]

    Notes
    -----
    Because all columns are Astropy Quantities (see
    `~astropy.units.Quantity`), one can easily convert between linear and
    log-scale at a later stage. the `scale` parameter just defines the
    scale to use initially.
    '''

    modes = ['continuum', 'spectroscopy']
    scales = ['linear', 'dB']

    assert mode in modes, 'mode must be either "continuum" or "spectroscopy"'
    assert scale in scales, 'scale must be either "linear" or "dB"'
    midx = modes.index(mode)

    csv_name = 'ra_769_table{}_limits_{}.csv'.format(midx + 1, mode)
    this_dir, this_filename = os.path.split(__file__)
    csv_path = os.path.join(this_dir, '../itudata/ra.769-2', csv_name)

    csv_tab = np.genfromtxt(
        csv_path, delimiter=',', skip_header=1, names=True, dtype=np.float64
        )

    qtab = QTable(meta={'name': 'RA.769 {} limits'.format(mode)})
    qtab['frequency'] = csv_tab['freq0'] * apu.MHz
    qtab['bandwidth'] = csv_tab['Delta_f'] * apu.MHz
    qtab['T_A'] = csv_tab['T_A'] * apu.K
    qtab['T_rx'] = csv_tab['T_rx'] * apu.K

    qtab['T_rms'] = (
        (qtab['T_A'] + qtab['T_rx']) /
        np.sqrt(2000. * apu.s * qtab['bandwidth'])
        ).to(apu.mK)
    P_rms_nu = con.k_B * qtab['T_rms']
    qtab['P_rms_nu'] = P_rms_nu.to(
        apu.Watt / apu.Hz if scale == 'linear' else cnv.dB_W_Hz
        )
    Plim = 0.1 * P_rms_nu * qtab['bandwidth']
    qtab['Plim'] = Plim.to(apu.Watt if scale == 'linear' else cnv.dB_W)
    qtab['Plim_nu'] = (Plim / qtab['bandwidth']).to(
        apu.W / apu.Hz if scale == 'linear' else cnv.dB_W_Hz
        )
    Slim = 4. * np.pi / con.c ** 2 * qtab['frequency'] ** 2 * Plim
    qtab['Slim'] = Slim.to(
        (apu.Watt / apu.m ** 2) if scale == 'linear' else cnv.dB_W_m2
        )
    qtab['Slim_nu'] = (Slim / qtab['bandwidth']).to(
        apu.Jy if scale == 'linear' else cnv.dB_W_m2_Hz
        )

    # field strength from full bandwidth:
    qtab['Efield'] = (np.sqrt(cnv.R0 * Slim)).to(
        (apu.microvolt / apu.m) if scale == 'linear' else cnv.dB_uV_m
        )
    # qtab['Efield2'] = (qtab['Efield'] ** 2).to(dB_uV_m)

    # now normalize for 1 MHz BW for comparison with spectroscopy
    # TODO: is this correct???
    qtab['Efield_norm'] = np.sqrt(
        cnv.R0 * Slim * apu.MHz / qtab['bandwidth'].to(apu.MHz)
        ).to(
        (apu.microvolt / apu.m) if scale == 'linear' else cnv.dB_uV_m
        )
    # qtab['Efield2_norm'] = (qtab['Efield_norm'] ** 2).to(dB_uV_m)

    # table formatting doesn't seem to work for QTable, so we convert
    tab = Table(qtab)

    for col in ['frequency', 'bandwidth', 'T_A', 'T_rx']:
        tab[col].format = '%.0f'

    for col in ['T_rms']:
        tab[col].format = '%.3f'

    for col in [
            'P_rms_nu',
            'Plim', 'Plim_nu', 'Slim', 'Slim_nu',
            'Efield', 'Efield_norm'
            ]:
        tab[col].format = '%.1e' if scale == 'linear' else '%.1f'

    return tab



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function
# from __future__ import unicode_literals

import pytest
import numpy as np
from numpy.testing import assert_equal, assert_allclose
from astropy.tests.helper import assert_quantity_allclose
from astropy import units as apu
from astropy.units import Quantity
from ... import conversions as cnv
from ...utils import check_astro_quantities
# from astropy.utils.misc import NumpyRNGContext


class TestConversions:

    def setup(self):

        pass

    def teardown(self):

        pass

    def test_db_scales(self):

        for o in [-10, -5, -1, 0, 1, 5, 10]:

            _db = 10 * o
            _val = 10 ** o

            assert_quantity_allclose(
                (_db * cnv.dB).to(cnv.dimless).value,
                _val
                )

            assert_quantity_allclose(
                (_db * cnv.dB_W).to(apu.W).value,
                _val
                )

            assert_quantity_allclose(
                (_db * cnv.dB_mW).to(apu.mW).value,
                _val
                )

            assert_quantity_allclose(
                (_db * cnv.dB_W_Hz).to(apu.W / apu.Hz).value,
                _val
                )

            assert_quantity_allclose(
                (_db * cnv.dB_W_m2).to(apu.W / apu.m ** 2).value,
                _val
                )

            assert_quantity_allclose(
                (_db * cnv.dB_W_m2_Hz).to(apu.W / apu.m ** 2 / apu.Hz).value,
                _val
                )

            assert_quantity_allclose(
                (_db * cnv.dB_Jy_Hz).to(apu.Jy * apu.Hz).value,
                _val
                )

            assert_quantity_allclose(
                (_db * cnv.dB_uV_m).to((apu.uV / apu.m) ** 2).value,
                _val
                )

    def test_constants(self):

        assert_quantity_allclose(
            cnv.R0, 376.7303135 * apu.Ohm
            )
        assert_quantity_allclose(
            cnv.Erx_unit, 5475.330657 * apu.uV / apu.m
            )

        # also check, if default unit is the most simple one:

        assert_equal(
            cnv.R0.unit, apu.Ohm
            )
        assert_equal(
            cnv.Erx_unit.unit, apu.uV / apu.m
            )

    def test_eff_from_geom_area(self):

        args_list = [
            (0, None, apu.m ** 2),
            (0, 100, apu.percent),
            ]
        check_astro_quantities(cnv.eff_from_geom_area, args_list)

        Aeff = 50. * apu.m ** 2
        Ageom = 100. * apu.m ** 2

        assert_quantity_allclose(
            cnv.eff_from_geom_area(Ageom, 50 * apu.percent),
            Aeff
            )
        assert_quantity_allclose(
            cnv.eff_from_geom_area(Ageom, 0.5 * cnv.dimless),
            Aeff
            )

    def test_geom_from_eff_area(self):

        args_list = [
            (0, None, apu.m ** 2),
            (0, 100, apu.percent),
            ]
        check_astro_quantities(cnv.geom_from_eff_area, args_list)

        Aeff = 50. * apu.m ** 2
        Ageom = 100. * apu.m ** 2

        assert_quantity_allclose(
            cnv.geom_from_eff_area(Aeff, 50 * apu.percent),
            Ageom
            )
        assert_quantity_allclose(
            cnv.geom_from_eff_area(Aeff, 0.5 * cnv.dimless),
            Ageom
            )

    def eta_a_from_areas(self):

        args_list = [
            (0, None, apu.m ** 2),
            (0, None, apu.m ** 2),
            ]
        check_astro_quantities(cnv.geom_from_eff_area, args_list)

        Aeff = 50. * apu.m ** 2
        Ageom = 100. * apu.m ** 2

        assert_quantity_allclose(
            cnv.geom_from_eff_area(Ageom, Aeff),
            50 * apu.percent
            )

    def test_gain_from_eff_area(self):

        args_list = [
            (0, None, apu.m ** 2),
            (0, None, apu.Hz),
            ]
        check_astro_quantities(cnv.gain_from_eff_area, args_list)

        assert_quantity_allclose(
            cnv.gain_from_eff_area(50. * apu.m ** 2, 1 * apu.GHz),
            38.4453846250226 * cnv.dBi
            )

        assert_quantity_allclose(
            cnv.gain_from_eff_area(50. * apu.m ** 2, 10 * apu.GHz),
            58.4453846250226 * cnv.dBi
            )

    def test_eff_area_from_gain(self):

        args_list = [
            (1.e-30, None, cnv.dimless),
            (0, None, apu.Hz),
            ]
        check_astro_quantities(cnv.eff_area_from_gain, args_list)

        assert_quantity_allclose(
            cnv.eff_area_from_gain(38.4453846250226 * cnv.dB, 1 * apu.GHz),
            50. * apu.m ** 2
            )
        assert_quantity_allclose(
            cnv.eff_area_from_gain(58.4453846250226 * cnv.dB, 10 * apu.GHz),
            50. * apu.m ** 2
            )

    def test_powerflux_from_efield(self):

        args_list = [
            (1.e-30, None, apu.V / apu.meter),
            ]
        check_astro_quantities(cnv.powerflux_from_efield, args_list)

        assert_quantity_allclose(
            cnv.powerflux_from_efield(20 * cnv.dB_uV_m),
            2.654418729345079e-13 * apu.W / apu.m ** 2
            )
        assert_quantity_allclose(
            cnv.powerflux_from_efield(10 * apu.uV / apu.m),
            2.654418729345079e-13 * apu.W / apu.m ** 2
            )

        assert_quantity_allclose(
            cnv.powerflux_from_efield(100 * apu.uV ** 2 / apu.m ** 2),
            2.654418729345079e-13 * apu.W / apu.m ** 2
            )

    def test_efield_from_powerflux(self):

        args_list = [
            (None, None, cnv.dB_W_m2),
            ]
        check_astro_quantities(cnv.efield_from_powerflux, args_list)

        assert_quantity_allclose(
            cnv.efield_from_powerflux(10 * cnv.dB_W_m2).to(cnv.dB_uV_m),
            155.76030566965238 * cnv.dB_uV_m
            )
        assert_quantity_allclose(
            cnv.efield_from_powerflux(10 * apu.W / apu.m ** 2),
            61378360.4762272 * apu.uV / apu.m
            )

    def test_ptx_from_efield(self):

        args_list = [
            (1.e-30, None, apu.V / apu.meter),
            (1.e-30, None, apu.m),
            (1.e-30, None, cnv.dimless),
            ]
        check_astro_quantities(cnv.ptx_from_efield, args_list)

        assert_quantity_allclose(
            cnv.ptx_from_efield(20 * cnv.dB_uV_m, 1 * apu.km, 20 * cnv.dBi),
            3.3356409518646604e-08 * apu.W
            )
        assert_quantity_allclose(
            cnv.ptx_from_efield(10 * apu.uV / apu.m, 1 * apu.km, 20 * cnv.dBi),
            3.3356409518646604e-08 * apu.W
            )

        assert_quantity_allclose(
            cnv.ptx_from_efield(
                100 * apu.uV ** 2 / apu.m ** 2, 1 * apu.km, 20 * cnv.dBi
                ),
            3.3356409518646604e-08 * apu.W
            )

    def test_efield_from_ptx(self):

        args_list = [
            (1.e-30, None, apu.W),
            (1.e-30, None, apu.m),
            (1.e-30, None, cnv.dimless),
            ]
        check_astro_quantities(cnv.efield_from_ptx, args_list)

        assert_quantity_allclose(
            cnv.efield_from_ptx(10 * cnv.dB_W, 1 * apu.km, 20 * cnv.dBi),
            173145.15817963346 * apu.uV / apu.meter
            )
        assert_quantity_allclose(
            cnv.efield_from_ptx(10 * apu.W, 1 * apu.km, 20 * cnv.dBi),
            173145.15817963346 * apu.uV / apu.meter
            )

    def test_powerflux_from_ptx(self):

        args_list = [
            (1.e-30, None, apu.W),
            (1.e-30, None, apu.m),
            (1.e-30, None, cnv.dimless),
            ]
        check_astro_quantities(cnv.powerflux_from_ptx, args_list)

        assert_quantity_allclose(
            cnv.powerflux_from_ptx(10 * cnv.dB_W, 1 * apu.km, 20 * cnv.dBi),
            7.957747154594768e-05 * apu.W / apu.m ** 2
            )
        assert_quantity_allclose(
            cnv.powerflux_from_ptx(10 * apu.W, 1 * apu.km, 20 * cnv.dBi),
            7.957747154594768e-05 * apu.W / apu.m ** 2
            )

    def test_ptx_from_powerflux(self):

        args_list = [
            (1.e-30, None, apu.W / apu.m ** 2),
            (1.e-30, None, apu.m),
            (1.e-30, None, cnv.dimless),
            ]
        check_astro_quantities(cnv.ptx_from_powerflux, args_list)

        assert_quantity_allclose(
            cnv.ptx_from_powerflux(10 * cnv.dB_W_m2, 1 * apu.km, 20 * cnv.dBi),
            1256637.0614359172 * apu.W
            )
        assert_quantity_allclose(
            cnv.ptx_from_powerflux(
                10 * apu.W / apu.m ** 2, 1 * apu.km, 20 * cnv.dBi
                ),
            1256637.0614359172 * apu.W
            )

    def test_powerflux_from_prx(self):

        args_list = [
            (1.e-30, None, apu.W),
            (1.e-30, None, apu.Hz),
            (1.e-30, None, cnv.dimless),
            ]
        check_astro_quantities(cnv.powerflux_from_prx, args_list)

        assert_quantity_allclose(
            cnv.powerflux_from_prx(10 * cnv.dB_W, 1 * apu.GHz, 20 * cnv.dBi),
            13.981972968457278 * apu.W / apu.m ** 2
            )
        assert_quantity_allclose(
            cnv.powerflux_from_prx(10 * apu.W, 1 * apu.GHz, 20 * cnv.dBi),
            13.981972968457278 * apu.W / apu.m ** 2
            )

    def test_prx_from_powerflux(self):

        args_list = [
            (1.e-30, None, apu.W / apu.m ** 2),
            (1.e-30, None, apu.Hz),
            (1.e-30, None, cnv.dimless),
            ]
        check_astro_quantities(cnv.prx_from_powerflux, args_list)

        assert_quantity_allclose(
            cnv.prx_from_powerflux(
                -160 * cnv.dB_W_m2, 1 * apu.Hz, 20 * cnv.dBi
                ),
            71.52066466270222 * apu.W
            )
        assert_quantity_allclose(
            cnv.prx_from_powerflux(
                1E-16 * apu.W / apu.m ** 2, 1 * apu.Hz, 20 * cnv.dBi
                ),
            71.52066466270222 * apu.W
            )

    def test_free_space_loss(self):

        args_list = [
            (1.e-30, None, apu.m),
            (1.e-30, None, apu.Hz),
            ]
        check_astro_quantities(cnv.free_space_loss, args_list)

        assert_quantity_allclose(
            cnv.free_space_loss(1 * apu.km, 1 * apu.GHz),
            -92.44778322188337 * cnv.dB
            )

    def test_prx_from_ptx(self):

        args_list = [
            (1.e-30, None, apu.W),
            (1.e-30, None, cnv.dimless),
            (1.e-30, None, cnv.dimless),
            (1.e-30, None, apu.m),
            (1.e-30, None, apu.Hz),
            ]
        check_astro_quantities(cnv.prx_from_ptx, args_list)

        assert_quantity_allclose(
            cnv.prx_from_ptx(
                10 * cnv.dB_W,
                50 * cnv.dBi, 50 * cnv.dBi,
                1 * apu.km, 1 * apu.GHz
                ),
            56.9143365714346 * apu.W
            )
        assert_quantity_allclose(
            cnv.prx_from_ptx(
                10 * apu.W,
                50 * cnv.dBi, 50 * cnv.dBi,
                1 * apu.km, 1 * apu.GHz
                ),
            56.9143365714346 * apu.W
            )

    def test_ptx_from_prx(self):

        args_list = [
            (1.e-30, None, apu.W),
            (1.e-30, None, cnv.dimless),
            (1.e-30, None, cnv.dimless),
            (1.e-30, None, apu.m),
            (1.e-30, None, apu.Hz),
            ]
        check_astro_quantities(cnv.ptx_from_prx, args_list)

        assert_quantity_allclose(
            cnv.ptx_from_prx(
                10 * cnv.dB_W,
                50 * cnv.dBi, 50 * cnv.dBi,
                1 * apu.km, 1 * apu.GHz
                ),
            1.7570265424158553 * apu.W
            )
        assert_quantity_allclose(
            cnv.ptx_from_prx(
                10 * apu.W,
                50 * cnv.dBi, 50 * cnv.dBi,
                1 * apu.km, 1 * apu.GHz
                ),
            1.7570265424158553 * apu.W
            )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

""" This module contains helper functions for accessing, downloading, and
caching data files.
"""

import os
from ..utils.introspection import find_current_module, resolve_name

try:
    import pathlib
except ImportError:
    HAS_PATHLIB = False
else:
    HAS_PATHLIB = True

__all__ = ['get_pkg_data_filename']


def get_pkg_data_filename(data_name, package=None):
    """
    Retrieves a data file from the standard locations for the package and
    provides a local filename for the data.

    Parameters
    ----------
    data_name : str
        Name/location of the desired data file.  One of the following:

            * The name of a data file included in the source
              distribution.  The path is relative to the module
              calling this function.  For example, if calling from
              ``astropy.pkname``, use ``'data/file.dat'`` to get the
              file in ``astropy/pkgname/data/file.dat``.  Double-dots
              can be used to go up a level.  In the same example, use
              ``'../data/file.dat'`` to get ``astropy/data/file.dat``.

    package : str, optional
        If specified, look for a file relative to the given package, rather
        than the default of looking relative to the calling module's package.

    Raises
    ------
    IOError
        If problems occur writing or reading a local file.

    Returns
    -------
    filename : str
        A file path on the local file system corresponding to the data
        requested in ``data_name``.

    Examples
    --------

    This will retrieve the contents of the data file for the `astropy.wcs`
    tests::

        >>> from astropy.utils.data import get_pkg_data_filename
        >>> fn = get_pkg_data_filename('data/3d_cd.hdr',
        ...                            package='astropy.wcs.tests')
        >>> with open(fn) as f:
        ...     fcontents = f.read()
        ...

    """

    data_name = os.path.normpath(data_name)

    datafn = _find_pkg_data_path(data_name, package=package)
    if os.path.isdir(datafn):
        raise IOError("Tried to access a data file that's actually "
                      "a package data directory")
    elif os.path.isfile(datafn):  # local file
        return datafn
    else:
        raise IOError('File {} not found!'.format(datafn))


def _is_inside(path, parent_path):
    # We have to try realpath too to avoid issues with symlinks, but we leave
    # abspath because some systems like debian have the absolute path (with no
    # symlinks followed) match, but the real directories in different
    # locations, so need to try both cases.
    return os.path.abspath(path).startswith(os.path.abspath(parent_path)) \
        or os.path.realpath(path).startswith(os.path.realpath(parent_path))


def _find_pkg_data_path(data_name, package=None):
    """
    Look for data in the source-included data directories and return the
    path.
    """

    if package is None:
        module = find_current_module(1, True)

        if module is None:
            # not called from inside an astropy package.  So just pass name
            # through
            return data_name

        if not hasattr(module, '__package__') or not module.__package__:
            # The __package__ attribute may be missing or set to None; see
            # PEP-366, also astropy issue #1256
            if '.' in module.__name__:
                package = module.__name__.rpartition('.')[0]
            else:
                package = module.__name__
        else:
            package = module.__package__
    else:
        module = resolve_name(package)

    rootpkgname = package.partition('.')[0]

    rootpkg = resolve_name(rootpkgname)

    module_path = os.path.dirname(module.__file__)
    path = os.path.join(module_path, data_name)

    root_dir = os.path.dirname(rootpkg.__file__)
    if not _is_inside(path, root_dir):
        raise RuntimeError("attempted to get a local data file outside "
                           "of the {} tree.".format(rootpkgname))

    return path



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import (
    absolute_import, unicode_literals, division, print_function
    )

from copy import deepcopy
import pytest
from astropy import units as apu
import numpy as np
from .. import conversions as cnv


__all__ = [
    'check_astro_quantities',
    ]


def tiny_fraction(num, digit=6):
    '''
    Return a tiny fraction of a number (in the specified digit)
    '''

    num = np.abs(num)
    num = np.max([num, 1.e-30])
    return 10 ** (np.log10(num) - digit)


def check_astro_quantities(
        func, args_list, kwargs_list=None, invalid_unit=apu.byte
        ):

    if kwargs_list is None:
        kwargs_list = []

    def make_args(alist, klist):

        args = []
        kwargs = {}
        for lowval, hival, unit in args_list:

            if lowval is None and hival is not None:
                lowval = hival
            elif lowval is not None and hival is None:
                hival = lowval
            elif lowval is None and hival is None:
                lowval = hival = 0.

            args.append(np.mean((lowval, hival)) * unit)

        for var, lowval, hival, unit in kwargs_list:

            if lowval is None and hival is not None:
                lowval = hival
            elif lowval is not None and hival is None:
                hival = lowval
            elif lowval is None and hival is None:
                lowval = hival = 0.

            kwargs[var] = np.mean((lowval, hival)) * unit

        return args, kwargs

    args, kwargs = make_args(args_list, kwargs_list)

    for case, argtup in enumerate(args_list):

        # test for value ranges
        _args = deepcopy(args)

        inv_lowval = argtup[0]
        if inv_lowval is not None:
            inv_lowval -= tiny_fraction(inv_lowval)
            _args[case] = inv_lowval * _args[case].unit

            with pytest.raises(AssertionError):
                func(*_args, **kwargs)

        _args = deepcopy(args)

        inv_hival = argtup[1]
        if inv_hival is not None:
            inv_hival += tiny_fraction(inv_hival)
            _args[case] = inv_hival * _args[case].unit

            with pytest.raises(AssertionError):
                func(*_args, **kwargs)

        # test for wrong unit
        _args = deepcopy(args)

        _args[case] = _args[case].value * invalid_unit

        with pytest.raises(apu.UnitsError):
            func(*_args, **kwargs)

        # test for missing unit
        _args[case] = _args[case].value

        with pytest.raises(TypeError):
            func(*_args, **kwargs)

    for argtup in kwargs_list:

        # test for value ranges
        _kwargs = deepcopy(kwargs)
        var = argtup[0]

        inv_lowval = argtup[1]
        if inv_lowval is not None:
            inv_lowval -= tiny_fraction(inv_lowval)
            _kwargs[var] = inv_lowval * _kwargs[var].unit

            with pytest.raises(AssertionError):
                func(*args, **_kwargs)

        _kwargs = deepcopy(kwargs)

        inv_hival = argtup[2]
        if inv_hival is not None:
            inv_hival += tiny_fraction(inv_hival)
            _kwargs[var] = inv_hival * _kwargs[var].unit

            with pytest.raises(AssertionError):
                func(*args, **_kwargs)

        # test for wrong unit
        _kwargs = deepcopy(kwargs)

        _kwargs[var] = _kwargs[var].value * invalid_unit

        with pytest.raises(apu.UnitsError):
            func(*args, **_kwargs)

        # test for missing unit
        _kwargs[var] = _kwargs[var].value

        with pytest.raises(TypeError):
            func(*args, **_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import os


BASEPATH = 'R-REC-P.452-16-201507'


def main():

    lons = np.genfromtxt(os.path.join(BASEPATH, 'LON.TXT'))
    lats = np.genfromtxt(os.path.join(BASEPATH, 'LAT.TXT'))
    dn50 = np.genfromtxt(os.path.join(BASEPATH, 'DN50.TXT'))
    n050 = np.genfromtxt(os.path.join(BASEPATH, 'N050.TXT'))


    # import matplotlib.pyplot as plt

    # plt.contourf(lons, lats, dn50, 128)
    # plt.show()

    # plt.contourf(lons, lats, n050, 128)
    # plt.show()

    save_kwargs = {
        'lons': lons.astype(np.float32),
        'lats': lats.astype(np.float32),
        'dn50': dn50.astype(np.float32),
        'n050': n050.astype(np.float32),
        }
    np.savez('refract_map', **save_kwargs)

    # read using:
    # dat = np.load('refract_map.npz')


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
The atm subpackage provides an implementation of the atmospheric models of
`ITU-R Rec. P.676-10 <https://www.itu.int/rec/R-REC-P.676-10-201309-S/en>`_.
For this, various other algorithms from the following two ITU-R
Recommendations are
necessary:

    - `ITU-R Rec. P.835-5 <https://www.itu.int/rec/R-REC-P.835-5-201202-I/en>`_,
      which contains the Standard and several special atmospheric profiles.
    - `ITU-R Rec. P.453-12 <https://www.itu.int/rec/R-REC-P.453-12-201609-I/en>`_,
      which is used to calculate the refractive index from temperature and
      water/total air pressure. Furthermore, P.453 has formulae to derive the
      saturation water pressure from temperature and total air pressure,
      as well as the water pressure from temperature, pressure and humidity, or
      alternatively from temperature and wator vapor density.

Notes
-----
A new version of ITU-R P.676 is available,
`ITU-R Rec. P.676-11 <https://www.itu.int/rec/R-REC-P.676-11-201609-I/en>`_,
which has a modified algorithm for Annex 2. We will implement this soon.
'''

from .atm import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function
# from __future__ import unicode_literals

import pytest
from functools import partial
import numpy as np
from numpy.testing import assert_equal, assert_allclose
from astropy.tests.helper import assert_quantity_allclose
from astropy import units as apu
from astropy.units import Quantity
from ... import conversions as cnv
from ... import pathprof
from ...utils import check_astro_quantities
from astropy.utils.misc import NumpyRNGContext
from geographiclib.geodesic import Geodesic


TOL_KWARGS = {'atol': 1.e-4, 'rtol': 1.e-4}


class TestHelper:

    def setup(self):

        pass

    def teardown(self):

        pass

    def test_annual_timepercent_from_worst_month(self):

        pfunc = pathprof.annual_timepercent_from_worst_month
        args_list = [
            (0, 100, apu.percent),
            (-90, 90, apu.deg),
            (0, 100, apu.percent),
            ]
        check_astro_quantities(pfunc, args_list)

        p_w = Quantity([0.01, 10., 20., 20., 50., 50.], apu.percent)
        phi = Quantity([30., 50., -50., 50, 40., 40.], apu.deg)
        omega = Quantity([0., 0., 10., 10., 0., 100.], apu.percent)

        p = Quantity(
            [
                1.407780e-03, 4.208322e+00, 9.141926e+00,
                9.141926e+00, 4.229364e+01, 1.889433e+01
                ], apu.percent
            )
        assert_quantity_allclose(
            pfunc(p_w, phi, omega),
            p,
            **TOL_KWARGS
            )

    def test_radiomet_data_for_pathcenter(self):

        pfunc = pathprof.radiomet_data_for_pathcenter
        args_list = [
            (0, 360, apu.deg),
            (-90, 90, apu.deg),
            (0, None, apu.km),
            (0, None, apu.km),
            ]
        check_astro_quantities(pfunc, args_list)

        lon = Quantity([0.01, 10., 20., 20., 50., 359.], apu.deg)
        lat = Quantity([30., 50., -50., 50, 40., 40.], apu.deg)
        d_tm = Quantity([0.1, 1., 10., 10., 0., 100.], apu.km)
        d_lm = Quantity([0.1, 1., 10., 10., 0., 100.], apu.km)

        DN, beta_0, N0 = pfunc(lon, lat, d_tm, d_lm)

        print(DN)
        print(beta_0)
        print(N0)

        assert_quantity_allclose(
            DN,
            Quantity([
                33.82150749, 36.5930002, 44.36488936, 35.57166629,
                40.73833338, 44.74622175], cnv.dimless / apu.km),
            )

        assert_quantity_allclose(
            beta_0,
            Quantity([
                16.57415849, 8.10024667, 6.27440928, 6.27440928,
                11.74897555, 2.60825487], apu.percent),
            )

        assert_quantity_allclose(
            N0,
            Quantity([
                304.58966064, 323.69878472, 319.52155219, 322.11834378,
                326.14456177, 329.69577705], cnv.dimless),
            )

    def test_eff_earth_radius_median(self):

        pfunc = pathprof.eff_earth_radius_median
        args_list = [
            (0, 360, apu.deg),
            (-90, 90, apu.deg),
            ]
        check_astro_quantities(pfunc, args_list)

        lon = Quantity([0.01, 10., 20., 20., 50., 359.], apu.deg)
        lat = Quantity([30., 50., -50., 50, 40., 40.], apu.deg)

        a_e = pfunc(lon, lat)

        print(a_e)

        assert_quantity_allclose(
            a_e,
            Quantity([
                8120.30557961, 8307.21637166, 8880.41920755, 8237.34436165,
                8603.41184774, 8910.586491], apu.km),
            )



#!python
# -*- coding: utf-8 -*-
# cython: language_level=3
# cython: cdivision=True, boundscheck=False, wraparound=False
# cython: embedsignature=True

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

cimport cython
cimport numpy as np
from numpy cimport PyArray_MultiIter_DATA as Py_Iter_DATA
from libc.math cimport (
    exp, sqrt, fabs, M_PI, sin, cos, tan, asin, acos, atan2
    )
import numpy as np

np.import_array()

# __all__ = ['inverse', 'direct']


cdef double WGS_a = 6378137.0
cdef double WGS_b = 6356752.314245
cdef double WGS_f = 1 / 298.257223563
cdef double DEG2RAD = M_PI / 180.
cdef double RAD2DEG = 180. / M_PI
cdef double M_2PI = 2 * M_PI

# see https://en.wikipedia.org/wiki/Vincenty's_formulae

cdef (double, double, double) _inverse(
        double lon1_rad, double lat1_rad,
        double lon2_rad, double lat2_rad,
        double eps,
        int maxiter,
        ) nogil:

    cdef:
        # note: "a" is short for alpha, "s" for sigma in this function
        double U1, U2, L, lam, last_lam, a, s
        double sin_U1, cos_U1, sin_U2, cos_U2, tan_U1, tan_U2
        double sin_lam, cos_lam
        double sin_s, cos_s, sin_a, cos2_a, cos_2sm, cos2_2sm
        double C, u2, A, B, ds

        double dist, bearing1_rad, bearing2_rad

        int _iter

    tan_U1 = (1 - WGS_f) * tan(lat1_rad)
    tan_U2 = (1 - WGS_f) * tan(lat2_rad)
    L = lon2_rad - lon1_rad

    cos_U1 = 1. / sqrt(1 + tan_U1 ** 2)
    cos_U2 = 1. / sqrt(1 + tan_U2 ** 2)
    sin_U1 = tan_U1 * cos_U1
    sin_U2 = tan_U2 * cos_U2

    lam = last_lam = L
    _iter = 0

    while True:

        sin_lam = sin(lam)
        cos_lam = cos(lam)

        sin_s = sqrt(
            (cos_U2 * sin_lam) ** 2 +
            (cos_U1 * sin_U2 - sin_U1 * cos_U2 * cos_lam) ** 2
            )
        cos_s = sin_U1 * sin_U2 + cos_U1 * cos_U2 * cos_lam
        s = atan2(sin_s, cos_s)
        sin_a = cos_U1 * cos_U2 * sin_lam / sin_s
        cos2_a = 1 - sin_a ** 2
        cos_2sm = cos_s - 2 * sin_U1 * sin_U2 / cos2_a
        cos2_2sm = cos_2sm ** 2

        C = WGS_f / 16 * cos2_a * (4 + WGS_f * (4 - 3 * cos2_a))
        lam = L + (1 - C) * WGS_f * sin_a * (
            s + C * sin_s * (
                cos_2sm + C * cos_s * (-1 + 2 * cos2_2sm)
                )
            )

        if fabs(lam - last_lam) < eps or _iter > maxiter:
            break

        _iter += 1
        last_lam = lam

    u2 = cos2_a * (WGS_a ** 2 - WGS_b ** 2) / WGS_b ** 2
    A = 1 + u2 / 16384 * (4096 + u2 * (-768 + u2 * (320 - 175 * u2)))
    B = u2 / 1024 * (256 + u2 * (-128 + u2 * (74 - 47 * u2)))

    ds = B * sin_s * (
        cos_2sm + 0.25 * B * (
            cos_s * (-1 + 2 * cos2_2sm) -
            1. / 6. * B * cos_2sm * (-3 + 4 * sin_s ** 2) * (-3 + 4 * cos2_2sm)
            )
        )

    dist = WGS_b * A * (s - ds)
    bearing1_rad = atan2(
        cos_U2 * sin_lam,
        cos_U1 * sin_U2 - sin_U1 * cos_U2 * cos_lam
        )
    bearing2_rad = atan2(
        cos_U1 * sin_lam,
        -sin_U1 * cos_U2 + cos_U1 * sin_U2 * cos_lam
        )

    return (dist, bearing1_rad, bearing2_rad)


def inverse_cython(
        lon1_rad, lat1_rad,
        lon2_rad, lat2_rad,
        double eps=1.e-12,  # corresponds to approximately 0.06mm
        int maxiter=50,
        ):

    cdef:

        np.broadcast broadcast, it
        np.ndarray dist_arr, bearing1_arr, bearing2_arr

    # Turn all inputs into arrays
    kwargs = dict(dtype=np.double, order='C', copy=False, subok=True)
    lon1_rad = np.array(lon1_rad, **kwargs)
    lat1_rad = np.array(lat1_rad, **kwargs)
    lon2_rad = np.array(lon2_rad, **kwargs)
    lat2_rad = np.array(lat2_rad, **kwargs)

    broadcast = np.broadcast(lon1_rad, lat1_rad, lon2_rad, lat2_rad)
    dist_arr = np.empty(broadcast.shape + (), dtype=np.double)
    bearing1_arr = np.empty(broadcast.shape + (), dtype=np.double)
    bearing2_arr = np.empty(broadcast.shape + (), dtype=np.double)

    it = np.broadcast(
        lon1_rad, lat1_rad, lon2_rad, lat2_rad,
        dist_arr, bearing1_arr, bearing2_arr,
        )

    with nogil:
        while np.PyArray_MultiIter_NOTDONE(it):

            (
                (<double*> Py_Iter_DATA(it, 4))[0],
                (<double*> Py_Iter_DATA(it, 5))[0],
                (<double*> Py_Iter_DATA(it, 6))[0],
                ) = _inverse(
                    (<double*> Py_Iter_DATA(it, 0))[0],
                    (<double*> Py_Iter_DATA(it, 1))[0],
                    (<double*> Py_Iter_DATA(it, 2))[0],
                    (<double*> Py_Iter_DATA(it, 3))[0],
                    eps,
                    maxiter,
                    )

            np.PyArray_MultiIter_NEXT(it)

    return dist_arr, bearing1_arr, bearing2_arr


cdef (double, double, double) _direct(
        double lon1_rad, double lat1_rad,
        double bearing1_rad,
        double dist,
        double eps,
        int maxiter,
        int cwrap
        ) nogil:

    cdef:
        # note: "a" is short for alpha, "s" for sigma in this function
        double U1, L, lam, a, s, s1, last_s
        double sin_U1, cos_U1, tan_U1, sin_a1, cos_a1
        double sin_lam, cos_lam
        double sin_s, cos_s, sin_a, cos2_a, cos_2sm, cos2_2sm
        double C, u2, A, B, ds

        double lon2_rad, lat2_rad, bearing2_rad

        int _iter

    tan_U1 = (1 - WGS_f) * tan(lat1_rad)

    cos_U1 = 1. / sqrt(1 + tan_U1 ** 2)
    sin_U1 = tan_U1 * cos_U1
    sin_a1 = sin(bearing1_rad)
    cos_a1 = cos(bearing1_rad)

    s1 = atan2(tan_U1, cos_a1)

    sin_a = cos_U1 * sin_a1
    sin2_a = sin_a ** 2
    cos2_a = 1 - sin2_a

    u2 = cos2_a * (WGS_a ** 2 - WGS_b ** 2) / WGS_b ** 2
    A = 1 + u2 / 16384 * (4096 + u2 * (-768 + u2 * (320 - 175 * u2)))
    B = u2 / 1024 * (256 + u2 * (-128 + u2 * (74 - 47 * u2)))

    s = last_s = dist / WGS_b / A
    _iter = 0

    while True:

        cos_2sm = cos(2 * s1 + s)
        cos2_2sm = cos_2sm ** 2

        sin_s = sin(s)
        cos_s = cos(s)

        ds = B * sin_s * (
            cos_2sm + 0.25 * B * (
                cos_s * (-1 + 2 * cos2_2sm) -
                1. / 6. * B * cos_2sm * (-3 + 4 * sin_s ** 2) *
                (-3 + 4 * cos2_2sm)
                )
            )

        s = dist / WGS_b / A + ds

        if fabs(s - last_s) < eps or _iter > maxiter:
            break

        _iter += 1
        last_s = s

    lat2_rad = atan2(
        sin_U1 * cos_s + cos_U1 * sin_s * cos_a1,
        (1 - WGS_f) * sqrt(
            sin2_a + (sin_U1 * sin_s - cos_U1 * cos_s * cos_a1) ** 2
            )
        )

    lam = atan2(
        sin_s * sin_a1,
        cos_U1 * cos_s - sin_U1 * sin_s * cos_a1
        )

    C = WGS_f / 16 * cos2_a * (4 + WGS_f * (4 - 3 * cos2_a))
    L = lam - (1 - C) * WGS_f * sin_a * (
        s + C * sin_s * (
            cos_2sm + C * cos_s * (-1 + 2 * cos2_2sm)
            )
        )

    lon2_rad = L + lon1_rad
    bearing2_rad = atan2(
        sin_a,
        -sin_U1 * sin_s + cos_U1 * cos_s * cos_a1
        )

    if cwrap:
        lon2_rad = (lon2_rad + M_PI) % M_2PI - M_PI

    return lon2_rad, lat2_rad, bearing2_rad


def direct_cython(
        lon1_rad, lat1_rad,
        bearing1_rad, dist_m,
        double eps=1.e-12,  # corresponds to approximately 0.06mm
        int maxiter=50,
        wrap=True,
        ):

    cdef:

        np.broadcast broadcast, it
        np.ndarray lon2_arr, lat2_arr, bearing2_arr
        int cwrap = 1 if wrap else 0

    # Turn all inputs into arrays
    kwargs = dict(dtype=np.double, order='C', copy=False, subok=True)
    lon1_rad = np.array(lon1_rad, **kwargs)
    lat1_rad = np.array(lat1_rad, **kwargs)
    bearing1_rad = np.array(bearing1_rad, **kwargs)
    dist_m = np.array(dist_m, **kwargs)

    broadcast = np.broadcast(lon1_rad, lat1_rad, bearing1_rad, dist_m)
    lon2_arr = np.empty(broadcast.shape + (), dtype=np.double)
    lat2_arr = np.empty(broadcast.shape + (), dtype=np.double)
    bearing2_arr = np.empty(broadcast.shape + (), dtype=np.double)

    it = np.broadcast(
        lon1_rad, lat1_rad, bearing1_rad, dist_m,
        lon2_arr, lat2_arr, bearing2_arr,
        )

    with nogil:
        while np.PyArray_MultiIter_NOTDONE(it):

            (
                (<double*> Py_Iter_DATA(it, 4))[0],
                (<double*> Py_Iter_DATA(it, 5))[0],
                (<double*> Py_Iter_DATA(it, 6))[0],
                ) = _direct(
                    (<double*> Py_Iter_DATA(it, 0))[0],
                    (<double*> Py_Iter_DATA(it, 1))[0],
                    (<double*> Py_Iter_DATA(it, 2))[0],
                    (<double*> Py_Iter_DATA(it, 3))[0],
                    eps,
                    maxiter,
                    cwrap
                    )

            np.PyArray_MultiIter_NEXT(it)

    return lon2_arr, lat2_arr, bearing2_arr


cdef inline int find_in_ordered(
        cython.floating[:] x,
        cython.floating x0,
        ) nogil:
    '''
    Find index of x0 in ordered vector x.

    Using a divide-and-conquer style algorithm.
    '''

    cdef:
        int l, h, i
        int length = x.shape[0]

    l = 0
    h = length - 1

    while x[l] <= x0 and x[h] >= x0:

        i = l + <int> ((x0 - x[l]) / (x[h] - x[l]) * (h - l))

        if x[i] < x0:
            l = i + 1
        elif x[i] > x0:
            h = i - 1
        else:
            return i

    return i


cdef inline double gauss1d(double offset, double s):

    return exp(-0.5 * offset * offset / s / s)


def regrid1d_with_x(
        cython.floating[:] x not None,
        cython.floating[:] y not None,
        cython.floating[:] x_new not None,
        cython.floating[:] y_new not None,  # output
        cython.floating width,
        bint regular=False,
        bint ordered=True,
        ):
    '''
    Regrid an array of values, measured at support x to a new support x_new.

    Example code::

        >>> from pycraf.pathprof import regrid1d_with_x
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        >>> x = np.linspace(1., 0., 100)
        >>> x_new = np.linspace(0., 1., 1000)
        >>> y_ = np.random.normal(0., 1., 100)
        >>> y_[50] = 10

        >>> y = np.empty_like(x)
        >>> y_new = np.empty_like(x_new)

        >>> regrid1d_with_x(x[::-1], y_[::-1], x, y, 0.005)
        >>> regrid1d_with_x(x[::-1], y_[::-1], x_new, y_new, 0.005)

        >>> plt.close()
        >>> plt.plot(x, y_, 'k-')
        >>> plt.plot(x, y, 'r-')
        >>> plt.plot(x_new, y_new, 'b-')
        >>> plt.show()
    '''

    cdef:

        double this_x, kv, ssum, norm

        int i, j, s, e
        int length = x.size
        int length_new = x_new.size

        double dx = fabs(x[0] - x[length - 1]) / length

    assert x.size == y.size, 'x and y must have equal size'

    for i in range(length_new):

        this_x = x_new[i]

        # find optimal s, e (if in ordered mode)
        if regular:

            s = int((this_x - 5. * width) / dx - 0.5)
            e = int((this_x + 5. * width) / dx + 1.5)
            if s < 0:
                s = 0
            if e >= length:
                e = length

        elif ordered:
            s = find_in_ordered(x, this_x)
            e = s + 1

            while True:
                s -= 1
                if s < 1:
                    s = 0
                    break
                if fabs(x[s] - this_x) > 5. * width:
                    break

            while True:
                e += 1
                if e >= length:
                    e = length
                    break
                if fabs(x[e - 1] - this_x) > 5. * width:
                    break

        else:
            s = 0
            e = length

        # print(i, s, e)

        norm = 0.
        ssum = 0.
        for j in range(s, e):
            kv = gauss1d(x[j] - this_x, width)
            ssum += kv * y[j]
            norm += kv

        # print(i, s, e, norm, ssum)
        if fabs(norm) < 1.e-12:
            y_new[i] = 0.
        else:
            y_new[i] = ssum / norm



#!/usr/bin/env python
# -*- coding: utf-8 -*-


from __future__ import (
    absolute_import, unicode_literals, division, print_function
    )

import numpy as np
from astropy import units as apu
from astropy.units import Quantity, UnitsError
import astropy.constants as con
from .. import utils


UNITS = [
    'dimless',
    'dB',
    'dBi',
    'dB_W',
    'dB_W_Hz',
    'dB_W_m2',
    'dB_W_m2_Hz',
    'dB_Jy_Hz',
    'dB_mW', 'dBm',
    'dB_mW_MHz', 'dBm_MHz',  # this is often used in engineering (dBm/MHz)
    'dB_1_m',
    'dB_uV_m',
    ]

__all__ = [
    'eff_from_geom_area', 'geom_from_eff_area',
    'eta_a_from_areas',
    'gain_from_eff_area', 'eff_area_from_gain',
    'antfactor_from_gain', 'gain_from_antfactor',
    'powerflux_from_efield', 'efield_from_powerflux',
    'ptx_from_efield', 'efield_from_ptx',
    'powerflux_from_ptx', 'ptx_from_powerflux',
    'prx_from_powerflux', 'powerflux_from_prx',
    'prx_from_ptx', 'ptx_from_prx',
    'free_space_loss',
    'Erx_unit', 'R0', 'efield_equivalency',
    ] + UNITS


# define some useful dB-Scales
# dimless = apu.Unit(1)
dimless = apu.dimensionless_unscaled
dB = dBi = apu.dB(dimless)
dB_W = apu.dB(apu.W)
dB_W_Hz = apu.dB(apu.W / apu.Hz)
dB_W_m2 = apu.dB(apu.W / apu.m ** 2)
dB_W_m2_Hz = apu.dB(apu.W / apu.Hz / apu.m ** 2)
dB_Jy_Hz = apu.dB(apu.Jy * apu.Hz)
dBm = dB_mW = apu.dB(apu.mW)
dBm_MHz = dB_mW_MHz = apu.dB(apu.mW / apu.MHz)
dB_uV_m = apu.dB(apu.uV ** 2 / apu.m ** 2)
dB_1_m = apu.dB(1. / apu.m)  # for antenna factor

# Astropy.unit equivalency between linear and logscale field strength
# this is necessary, because the dB_uV_m is from E ** 2 (dB scale is power)
# one can make use of the equivalency in the .to() function, e.g.:
#     Erx_unit.to(cnv.dB_uV_m, equivalencies=efield_equivalency)
# this conflicts with apu.logarithmic():
# def efield_equivalency():
#     return [(
#         apu.uV / apu.m,
#         dB_uV_m,
#         lambda x: 10. * np.log10(x ** 2),
#         lambda x: np.sqrt(10 ** (x / 10.))
#         )]


def efield_equivalency():
    '''
    `~astropy.units` equivalency to handle log-scale E-field units.

    For electric fields, the Decibel scale is define via the amplitude
    of the field squared, :math:`{\\vert\\vec E\\vert}^2` which is
    proportional to the power.

    Returns
    -------
    equivalency : list
        The returned list contains one tuple with the equivalency.
    '''
    return [(
        apu.uV / apu.m,
        (apu.uV / apu.m) ** 2,
        lambda x: x ** 2,
        lambda x: x ** 0.5
        )]


# apu.add_enabled_equivalencies(apu.logarithmic())
apu.add_enabled_equivalencies(efield_equivalency())

# define some useful constants
R0 = (
    1. * (con.mu0 / con.eps0) ** 0.5
    ).to(apu.Ohm)
Erx_unit = (
    (1 * apu.W / 4. / np.pi * R0) ** 0.5 / (1 * apu.km)
    ).to(apu.uV / apu.m)
C_VALUE = con.c.to(apu.m / apu.s).value
R0_VALUE = R0.to(apu.Ohm).value
ERX_VALUE = Erx_unit.to(apu.V / apu.m).value


@utils.ranged_quantity_input(
    geom_area=(0, None, apu.m ** 2),
    eta_a=(0, 100, apu.percent),
    strip_input_units=True, output_unit=apu.m ** 2
    )
def eff_from_geom_area(geom_area, eta_a):
    '''
    Effective antenna area from geometric area, given antenna efficiency.

    The effective and geometric antenna areas are linked via the antenna
    efficiency:

        A_eff = eta_a * A_geom.

    Parameters
    ----------
    geom_area : `~astropy.units.Quantity`
        Geometric antenna area, A_geom [m**2]
    eta_a : `~astropy.units.Quantity`
        Antenna efficiency [%, dimless]

    Returns
    -------
    eff_area : `~astropy.units.Quantity`
        Effective antenna area, A_eff [m**2]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return geom_area * eta_a / 100.


@utils.ranged_quantity_input(
    eff_area=(0, None, apu.m ** 2),
    eta_a=(0, 100, apu.percent),
    strip_input_units=True, output_unit=apu.m ** 2
    )
def geom_from_eff_area(eff_area, eta_a):
    '''
    Geometric antenna area from effective area, given antenna efficiency.

    The effective and geometric antenna areas are linked via the antenna
    efficiency:

        A_eff = eta_a * A_geom.

    Parameters
    ----------
    eff_area : `~astropy.units.Quantity`
        Effective antenna area, A_eff [m**2]
    eta_a : `~astropy.units.Quantity`
        Antenna efficiency [%, dimless]

    Returns
    -------
    geom_area : `~astropy.units.Quantity`
        Geometric antenna area, A_geom [m**2]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return eff_area / eta_a * 100.


@utils.ranged_quantity_input(
    geom_area=(0, None, apu.m ** 2),
    eff_area=(0, None, apu.m ** 2),
    strip_input_units=True, output_unit=apu.percent
    )
def eta_a_from_areas(geom_area, eff_area):
    '''
    Antenna efficiency from geometric and effective antenna areas.

    The effective and geometric antenna areas are linked via the antenna
    efficiency:

        A_eff = eta_a * A_geom.

    Parameters
    ----------
    eff_area : `~astropy.units.Quantity`
        Effective antenna area, A_eff [m**2]
    geom_area : `~astropy.units.Quantity`
        Geometric antenna area, A_geom [m**2]

    Returns
    -------
    eta_a : `~astropy.units.Quantity`
        Antenna efficiency [%, dimless]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return eff_area / geom_area * 100.


@utils.ranged_quantity_input(
    eff_area=(0, None, apu.m ** 2),
    freq=(0, None, apu.Hz),
    strip_input_units=True, output_unit=dBi
    )
def gain_from_eff_area(eff_area, freq):
    '''
    Antenna gain from effective antenna area, given frequency.

    Parameters
    ----------
    eff_area : `~astropy.units.Quantity`
        Effective antenna area, A_eff [m**2]
    freq : `~astropy.units.Quantity`
        Frequency [Hz]

    Returns
    -------
    gain : `~astropy.units.Quantity`
        Antenna gain [dBi]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return 10 * np.log10(
        4. * np.pi * eff_area * (freq / C_VALUE) ** 2
        )


@utils.ranged_quantity_input(
    gain=(1.e-30, None, dimless),
    freq=(0, None, apu.Hz),
    strip_input_units=True, output_unit=apu.m ** 2
    )
def eff_area_from_gain(gain, freq):
    '''
    Effective antenna area from antenna gain, given frequency.

    Parameters
    ----------
    gain : `~astropy.units.Quantity`
        Antenna gain [dBi, dimless]
    freq : `~astropy.units.Quantity`
        Frequency [Hz]

    Returns
    -------
    eff_area : `~astropy.units.Quantity`
        Effective antenna area, A_eff [m**2]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return gain * (C_VALUE / freq) ** 2 / 4. / np.pi


@utils.ranged_quantity_input(
    gain=(1.e-30, None, dimless),
    freq=(0, None, apu.Hz),
    zi=(0, None, apu.Ohm),
    strip_input_units=True, output_unit=dB_1_m
    )
def antfactor_from_gain(gain, freq, zi):
    '''
    Antenna factor from antenna gain, given frequency and impedance.

    Parameters
    ----------
    gain : `~astropy.units.Quantity`
        Antenna gain [dBi, or dimless]
    freq : `~astropy.units.Quantity`
        Frequency [Hz]
    zi : `~astropy.units.Quantity`
        Receiver impedance, Zi [Ohm]

    Returns
    -------
    antfactor : `~astropy.units.Quantity`
        Antenna factor, Ka [dB(1/m)]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return 10 * np.log10(np.sqrt(
        4. * np.pi / gain * (freq / C_VALUE) ** 2 * R0_VALUE / zi
        ))


@utils.ranged_quantity_input(
    antfactor=(1.e-30, None, 1. / apu.m),
    freq=(0, None, apu.Hz),
    zi=(0, None, apu.Ohm),
    strip_input_units=True, output_unit=dBi
    )
def gain_from_antfactor(antfactor, freq, zi):
    '''
    Antenna gain from antenna factor, given frequency and impedance.

    Parameters
    ----------
    antfactor : `~astropy.units.Quantity`
        Antenna factor, Ka [1/m]
    freq : `~astropy.units.Quantity`
        Frequency [Hz]
    zi : `~astropy.units.Quantity`
        Receiver impedance, Zi [Ohm]

    Returns
    -------
    gain : `~astropy.units.Quantity`
        Antenna gain [dBi, or dimless]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return 10 * np.log10(
        4. * np.pi / antfactor ** 2 * (freq / C_VALUE) ** 2 * R0_VALUE / zi
        )


# @apu.quantity_input(E=dB_uV_m, equivalencies=efield_equivalency())
@utils.ranged_quantity_input(
    efield=(1.e-30, None, apu.V / apu.meter),
    strip_input_units=True, output_unit=apu.W / apu.m ** 2
    )
def powerflux_from_efield(efield):
    '''
    Power flux density from E-field strength.

    Parameters
    ----------
    efield : `~astropy.units.Quantity`
        E-field strength, E [uV/m]

    Returns
    -------
    powerflux : `~astropy.units.Quantity`
        Power flux density, S [dB_W_m2, W/m**2]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return efield ** 2 / R0_VALUE


@utils.ranged_quantity_input(
    powerflux=(None, None, apu.W / apu.m ** 2),
    strip_input_units=True, output_unit=apu.uV / apu.meter
    )
def efield_from_powerflux(powerflux):
    '''
    E-field strength from power flux density.

    Parameters
    ----------
    powerflux : `~astropy.units.Quantity`
        Power flux density, S [dB_W_m2 or W/m**2]

    Returns
    -------
    efield : `~astropy.units.Quantity`
        E-field strength, E [uV/m]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return np.sqrt(powerflux * R0_VALUE) * 1.e6


@utils.ranged_quantity_input(
    efield=(1.e-30, None, apu.V / apu.meter),
    dist=(1.e-30, None, apu.m),
    gtx=(1.e-30, None, dimless),
    strip_input_units=True, output_unit=apu.W
    )
def ptx_from_efield(efield, dist, gtx):
    '''
    Transmitter power from E-field strength measured at distance.

    Parameters
    ----------
    efield : `~astropy.units.Quantity`
        E-field strength, E [dB_uV_m, uV/m, (uV/m)**2]
    dist : `~astropy.units.Quantity`
        Distance to transmitter [m]
    gtx : `~astropy.units.Quantity`
        Gain of transmitter, Gtx [dBi, or dimless]

    Returns
    -------
    ptx : `~astropy.units.Quantity`
        Transmitter power, Ptx [W]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return 4. * np.pi * dist ** 2 / gtx * efield ** 2 / R0_VALUE


@utils.ranged_quantity_input(
    ptx=(1.e-30, None, apu.W),
    dist=(1.e-30, None, apu.m),
    gtx=(1.e-30, None, dimless),
    strip_input_units=True, output_unit=apu.uV / apu.meter
    )
def efield_from_ptx(ptx, dist, gtx):
    '''
    E-field strength measured at distance from transmitter power.

    Parameters
    ----------
    ptx : `~astropy.units.Quantity`
        Transmitter power, Ptx [dB_W, W]
    dist : `~astropy.units.Quantity`
        Distance to transmitter [m]
    gtx : `~astropy.units.Quantity`
        Gain of transmitter, Gtx [dBi, dimless]

    Returns
    -------
    efield : `~astropy.units.Quantity`
        E-field strength, E [uV/m]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return (ptx * gtx / 4. / np.pi * R0_VALUE) ** 0.5 / dist * 1.e6


@utils.ranged_quantity_input(
    ptx=(1.e-30, None, apu.W),
    dist=(1.e-30, None, apu.m),
    gtx=(1.e-30, None, dimless),
    strip_input_units=True, output_unit=apu.W / apu.m ** 2
    )
def powerflux_from_ptx(ptx, dist, gtx):
    '''
    Power flux density from transmitter power.

    Parameters
    ----------
    ptx : `~astropy.units.Quantity`
        Transmitter power, Ptx [dB_W, W]
    dist : `~astropy.units.Quantity`
        Distance to transmitter [m]
    gtx : `~astropy.units.Quantity`
        Gain of transmitter, Gtx [dBi, dimless]

    Returns
    -------
    powerflux : `~astropy.units.Quantity`
        Power flux density, S (at distance) [W/m**2]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return gtx * ptx / 4. / np.pi / dist ** 2


@utils.ranged_quantity_input(
    powerflux=(1.e-30, None, apu.W / apu.m ** 2),
    dist=(1.e-30, None, apu.m),
    gtx=(1.e-30, None, dimless),
    strip_input_units=True, output_unit=apu.W
    )
def ptx_from_powerflux(powerflux, dist, gtx):
    '''
    Transmitter power from power flux density.

    Parameters
    ----------
    powerflux : `~astropy.units.Quantity`
        Power flux density, S (at distance) [W/m**2, dB_W_m2]
    dist : `~astropy.units.Quantity`
        Distance to transmitter [m]
    gtx : `~astropy.units.Quantity`
        Gain of transmitter, Gtx [dBi, dimless]

    Returns
    -------
    ptx : `~astropy.units.Quantity`
        Transmitter power, Ptx [W]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return powerflux * 4. * np.pi * dist ** 2 / gtx


@utils.ranged_quantity_input(
    prx=(1.e-30, None, apu.W),
    freq=(1.e-30, None, apu.Hz),
    grx=(1.e-30, None, dimless),
    strip_input_units=True, output_unit=apu.W / apu.m ** 2
    )
def powerflux_from_prx(prx, freq, grx):
    '''
    Power flux density from received power.

    Power flux density and received power are linked via effective antenna
    area (which is propotional to receiving antenna gain).

    Parameters
    ----------
    prx : `~astropy.units.Quantity`
        Received power [dB_W, W]
    freq : `~astropy.units.Quantity`
        Frequency of radiation [Hz]
    grx : `~astropy.units.Quantity`
        Gain of receiver, Grx [dBi, dimless]

    Returns
    -------
    powerflux : `~astropy.units.Quantity`
        Power flux density [W/m**2]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return prx / grx * (
        4. * np.pi * freq ** 2 / C_VALUE ** 2
        )


@utils.ranged_quantity_input(
    powerflux=(1.e-30, None, apu.W / apu.m ** 2),
    freq=(1.e-30, None, apu.Hz),
    grx=(1.e-30, None, dimless),
    strip_input_units=True, output_unit=apu.W
    )
def prx_from_powerflux(powerflux, freq, grx):
    '''
    Received power from power flux density.

    Power flux density and received power are linked via effective antenna
    area (which is propotional to receiving antenna gain).

    Parameters
    ----------
    powerflux : `~astropy.units.Quantity`
        Power flux density [W/m**2, dB_W_m2]
    freq : `~astropy.units.Quantity`
        Frequency of radiation [Hz]
    grx : `~astropy.units.Quantity`
        Gain of receiver, Grx [dBi, dimless]

    Returns
    -------
    prx : `~astropy.units.Quantity`
        Received power [W]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return powerflux * grx * (
        C_VALUE ** 2 / 4. / np.pi / freq ** 2
        )


def _free_space_loss(d, f):

    return (C_VALUE / 4. / np.pi / f / d) ** 2


@utils.ranged_quantity_input(
    dist=(1.e-30, None, apu.m),
    freq=(1.e-30, None, apu.Hz),
    strip_input_units=True, output_unit=dB
    )
def free_space_loss(dist, freq):
    '''
    Free-space loss of a propagating radio wave.

    Parameters
    ----------
    dist : `~astropy.units.Quantity`
        Distance between transmitter and receiver [m]
    freq : `~astropy.units.Quantity`
        Frequency of radiation [Hz]

    Returns
    -------
    FSPL : `~astropy.units.Quantity`
        Free-space loss [dB]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return 10. * np.log10(_free_space_loss(freq, dist))


@utils.ranged_quantity_input(
    ptx=(1.e-30, None, apu.W),
    gtx=(1.e-30, None, dimless),
    grx=(1.e-30, None, dimless),
    dist=(1.e-30, None, apu.m),
    freq=(1.e-30, None, apu.Hz),
    strip_input_units=True, output_unit=apu.W
    )
def prx_from_ptx(ptx, gtx, grx, dist, freq):
    '''
    Received power from transmitted power.

    Parameters
    ----------
    ptx : `~astropy.units.Quantity`
        Transmitter power, Ptx [dB_W, W]
    gtx : `~astropy.units.Quantity`
        Gain of transmitter, Gtx [dBi, dimless]
    grx : `~astropy.units.Quantity`
        Gain of receiver, Grx [dBi, dimless]
    dist : `~astropy.units.Quantity`
        Distance between transmitter and receiver [m]
    freq : `~astropy.units.Quantity`
        Frequency of radiation [Hz]

    Returns
    -------
    prx : `~astropy.units.Quantity`
        Received power [W]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return ptx * gtx * grx * _free_space_loss(dist, freq)


@utils.ranged_quantity_input(
    prx=(1.e-30, None, apu.W),
    gtx=(1.e-30, None, dimless),
    grx=(1.e-30, None, dimless),
    dist=(1.e-30, None, apu.m),
    freq=(1.e-30, None, apu.Hz),
    strip_input_units=True, output_unit=apu.W
    )
def ptx_from_prx(prx, gtx, grx, dist, freq):
    '''
    Transmitted power from received power.

    Parameters
    ----------
    prx : `~astropy.units.Quantity`
        Received power, Prx [dB_W, W]
    gtx : `~astropy.units.Quantity`
        Gain of transmitter, Gtx [dBi, dimless]
    grx : `~astropy.units.Quantity`
        Gain of receiver, Grx [dBi, dimless]
    dist : `~astropy.units.Quantity`
        Distance between transmitter and receiver [m]
    freq : `~astropy.units.Quantity`
        Frequency of radiation [Hz]

    Returns
    -------
    ptx : `~astropy.units.Quantity`
        Transmitter power, Ptx [W]

    Notes
    -----
    Because all parameters/returned values are Astropy Quantities (see
    `~astropy.units.Quantity`), unit conversion is automatically performed.
    '''

    return prx / gtx / grx / _free_space_loss(dist, freq)


if __name__ == '__main__':
    print('This not a standalone python program! Use as module.')
